#ifndef CLOUD_SEGMENTATION_H
#define CLOUD_SEGMENTATION_H

#include "types.h"
#include "Ply.hpp"
#include "Visualization_Tools.h"

using std::vector;
using std::cout;
using std::cerr;
using std::endl;
using std::string;

using I::misc::Ply;


namespace RegionGrowing {

typedef unsigned char byte;
const double PI = 3.141592;

class Cloud_segmentation
{
	
	struct HPoint {
		Point_d position;
		Vector_3 normal;
		int primitive_index;
	};


public:

	vector<HPoint> HPS;
	std::vector < std::vector < int > > spherical_neighborhood;
	std::vector < std::vector < int > > plane_point_index;
	std::vector < Plane_3 > extracted_planes;
	double average_nearest_spacing=0.;
	double average_k_spacing=0.;
	double std_k_spacing = 0.;
	double alpha_thresh = 2.0;

bool load_points(string filename)
{
	cout<<"Loading ";
	cout <<filename;

	HPS.clear();
	Ply ply;

	if (!ply.open(filename)){ return false;}

	for (Ply::ElementsIterator it = ply.elements_begin(); it != ply.elements_end(); ++it){
		const Ply::Element& element = *it;

		if (element.name() != "vertex"){
			if (!ply.skip(element)){ ply.close(); return false;}
			continue;
		}

		/* Check the properties are what we expect */
		if ((element.property(0).name() != "x")||(element.property(1).name() != "y")||(element.property(2).name() != "z")){cerr << "Unexpected vertex properties in the PLY file" << endl; ply.close(); return false;}


		size_t num_vertices = element.count();
		HPS.resize(num_vertices);

		

		if (element.num_properties() == 6) {
			for (size_t i=0; i<num_vertices; i++){
				double x,y,z;
				double nx,ny,nz;

				if ((!ply.read(element.property(0), x))||(!ply.read(element.property(1), y))||(!ply.read(element.property(2), z))){cerr << "error while reading (pos) vertex " << i+1 << endl; ply.close(); return false;}	
				if ((!ply.read(element.property(3), nx))||(!ply.read(element.property(4), ny))||(!ply.read(element.property(5), nz))){cerr << "error while reading attribut " << i+1 << endl; ply.close(); return false; }

				HPS[i].position = Point_d(x,y,z);
				HPS[i].normal = Vector_3(nx,ny,nz);
			}
		}

		

		else {cerr << "Unexpected vertex properties in the PLY file" << endl; ply.close(); return false;}
	}

	ply.close();

	cout<<" ("<<HPS.size()<<" points)"<<endl;   
	return true;
}


bool Compute_Knearest_neighbors(int K, bool apply_threshold=true)
{
	cout<<"Computing the K-nearest neighbors";

	//1-Neighborhood computation and reset the attributes of the structure points
	spherical_neighborhood.clear();
	spherical_neighborhood.reserve(HPS.size());
	std::map<Point_d,int> map_indice_point;
	std::list<Point_d> list_points;
	
	for(int i=0;i<(int)HPS.size();i++){
		const Point_d pt=HPS[i].position;
		map_indice_point[pt]=i;
		list_points.push_back(pt);
	}

	Tree tree(list_points.begin(), list_points.end());
	average_nearest_spacing = 0;
	average_k_spacing = 0;
	std_k_spacing = 0;
	int n_spacing = 0;

	int ten_percent_of_points = HPS.size() / 10;

	for(size_t i=0;i<HPS.size();i++){
		if (i % ten_percent_of_points == 0) std::cout << ". " << std::flush;

		Point_d query=HPS[i].position;
		Neighbor_search search(tree, query, K+1);

		double dist_1st_neighbor = FLT_MAX;
		std::vector < int > index_of_neighbors;
		index_of_neighbors.reserve(K);

		int terms = 0;
		double sum_k_spacing = 0;
		for(Neighbor_search::iterator it = search.begin(); it != search.end(); ++it){
			std::map<Point_d,int>::iterator iter = map_indice_point.begin();
			iter= map_indice_point.find(it->first);
			if( iter != map_indice_point.end() && iter->second!=i ) {
				index_of_neighbors.push_back(iter->second);
				double d = sqrt((query - iter->first).squared_length());
				sum_k_spacing += d;
				if (d < dist_1st_neighbor) dist_1st_neighbor = d;
				++terms;
			}
		}
		average_k_spacing += sum_k_spacing/terms;
		std_k_spacing += (sum_k_spacing/terms)*(sum_k_spacing/terms);
		spherical_neighborhood.push_back(index_of_neighbors);
		if (terms > 0) {
			average_nearest_spacing += dist_1st_neighbor;
			n_spacing++;
		}
	}
	average_nearest_spacing /= std::max(n_spacing, 1);
	average_k_spacing /= HPS.size();
	std_k_spacing = sqrt(std_k_spacing/HPS.size() - average_k_spacing*average_k_spacing);

	// Filter spherical_neighborhood and remove points too far from the query
	if (apply_threshold) {
		for(size_t i=0;i<HPS.size();i++) {
			Point_d query=HPS[i].position;
			std::vector < int > index_of_neighbors_filtered;
			for (int neighbor_index : spherical_neighborhood[i]) {
				Point_d pt=HPS[neighbor_index].position;
				double distance = sqrt((query - pt).squared_length());
				if (distance <= alpha_thresh*average_k_spacing) index_of_neighbors_filtered.push_back(neighbor_index);
			}
			spherical_neighborhood[i] = index_of_neighbors_filtered;
		}
	}

	cout<<endl<<endl;
	return true;
}


void set_extrema()
{
	double x_min, y_min, z_min, x_max, y_max, z_max;
	x_min = y_min = z_min = FLT_MAX;
	x_max = y_max = z_max = -FLT_MAX;
	for (size_t i = 0 ; i < HPS.size() ; ++i) {
		const Point_d & pt = HPS[i].position;
		const double &x = pt.x(), &y = pt.y(), &z = pt.z();
		if (x < x_min) x_min = x;
		if (x > x_max) x_max = x;
		if (y < y_min) y_min = y;
		if (y > y_max) y_max = y;
		if (z < z_min) z_min = z;
		if (z > z_max) z_max = z;
	}
	
	std::cout << "X : [" << x_min << ", " << x_max << "]" << std::endl;
	std::cout << "Y : [" << y_min << ", " << y_max << "]" << std::endl;
	std::cout << "Z : [" << z_min << ", " << z_max << "]" << std::endl;
}


bool region_growing(float epsilon_ratio, float normal_thresh, int Nmin){

		cout<<"Region growing ";
		assert(average_nearest_spacing>0.);
		float dist_threshold = epsilon_ratio*average_nearest_spacing;

		//Initialization structures
		plane_point_index.clear();

		for(int i=0; i<(int)HPS.size();i++) HPS[i].primitive_index=-1;
		
		int class_index=-1;

		// for each point, if not inliers yet..
		int ten_percent_of_points=HPS.size()/10;
		
		for(int i=0; i<(int)HPS.size();i++){

			if(i%ten_percent_of_points==0) cout<<". ";

			if(HPS[i].primitive_index==-1){

				//update the index of primitive
				++class_index;
				HPS[i].primitive_index=class_index;

				int conti=0; 	//for accelerate least_square fitting 

				//characteristics of the seed
				Vector_3 normal_seed=HPS[i].normal;
				Point_d pt_seed=HPS[i].position;
				Plane_3 optimal_plane(pt_seed,normal_seed);

				//initialization containers
				std::vector < int > index_container; index_container.push_back(i);
				std::vector < int > index_container_former_ring; index_container_former_ring.push_back(i);
				std::list < int > index_container_current_ring;

				//propagation
				bool propagation=true;
				do{

					propagation=false;

					for(int k=0;k<(int)index_container_former_ring.size();k++){

						int point_index=index_container_former_ring[k];

						for(int it=0;it<(int)spherical_neighborhood[point_index].size();it++){

							int neighbor_index=spherical_neighborhood[point_index][it];
							
							if( HPS[neighbor_index].primitive_index==-1 ){
								
								Point_d neighbor=HPS[neighbor_index].position;
								Point_d neighbor_projection=optimal_plane.projection(neighbor); 
								double proj_distance = sqrt((neighbor - neighbor_projection).squared_length());
								//double proj_distance=sqrt(pow(neighbor.x()-neighbor_projection.x(),2)+pow(neighbor.y()-neighbor_projection.y(),2)+pow(neighbor.z()-neighbor_projection.z(),2));

								if( proj_distance<dist_threshold && abs( HPS[neighbor_index].normal * optimal_plane.orthogonal_vector() ) > normal_thresh) { 

									HPS[neighbor_index].primitive_index=class_index;
									propagation=true;
									index_container_current_ring.push_back(neighbor_index);
									conti++;

									if( (conti<50 && conti%10==0) || (conti>50 && conti%500==0)){
										std::list < Point_d > listp;
										for(int pm=0;pm<(int)index_container.size();pm++){
											int yyh=index_container[pm];
											Point_d ptza=HPS[yyh].position;
											listp.push_back(ptza);
										}

										if (listp.size() >= 3) {
											Plane_3 reajusted_plane;
											double average_distance=linear_least_squares_fitting_3(listp.begin(),listp.end(),reajusted_plane, CGAL::Dimension_tag<0>());
											optimal_plane=reajusted_plane;
										}
									}
								}
							}
						}
					}

					//update containers
					index_container_former_ring.clear();
					for(std::list < int >::iterator it = index_container_current_ring.begin(); it != index_container_current_ring.end(); ++it){
						index_container_former_ring.push_back(*it);
						index_container.push_back(*it);
					}
					index_container_current_ring.clear();

				}while(propagation);
				

				//Test the number of inliers -> reject if inferior to Nmin
				if(index_container.size()>=Nmin) plane_point_index.push_back(index_container);
				else{ 
					class_index--;
					HPS[i].primitive_index=-1;
					for(int k=0;k<(int)index_container.size();k++) HPS[index_container[k]].primitive_index=-1; 
				}
			} 
		}
		
		
		cout<<endl<<"-> "<<plane_point_index.size()<<" primitives"<<endl<<endl;
		cout<<endl;
		return true;
	}


protected: 


};
}

#endif 
