#include "distance_field.h"

DistanceField::DistanceField(Bbox *bbox)
    : _m_bbox(bbox)
    , m_cutting_plane_ready(false)
{   
    diag = FT(-1);
}

DistanceField::~DistanceField()
{
    clear();
}

void DistanceField::clear()
{   
    m_cutting_plane_ready = false;
    m_max_distance_function = FT(0);
    diag = FT(-1);
}

FT DistanceField::random_in(const double a,
                    const double b)
{
    double r = rand() / (double)RAND_MAX;
    return (FT)(a + (b - a) * r);
}

Point3 DistanceField::random_point(const CGAL::Bbox_3& bbox)
{
    FT x = random_in(bbox.xmin(),bbox.xmax());
    FT y = random_in(bbox.ymin(),bbox.ymax());
    FT z = random_in(bbox.zmin(),bbox.zmax());
    return Point(x,y,z);
}

Vector DistanceField::random_vector()
{
    FT x = random_in(0.0,1.0);
    FT y = random_in(0.0,1.0);
    FT z = random_in(0.0,1.0);
    return Vector(x,y,z);
}

Ray DistanceField::random_ray(const CGAL::Bbox_3& bbox)
{
    Point p = random_point(bbox);
    Point q = random_point(bbox);
    return Ray(p,q);
}

Segment DistanceField::random_segment(const CGAL::Bbox_3& bbox)
{
    Point p = random_point(bbox);
    Point q = random_point(bbox);
    return Segment(p,q);
}

Line DistanceField::random_line(const CGAL::Bbox_3& bbox)
{
    Point p = random_point(bbox);
    Point q = random_point(bbox);
    return Line(p,q);
}

Plane DistanceField::random_plane(const CGAL::Bbox_3& bbox)
{
    Point p = random_point(bbox);
    Vector vec = random_vector();
    return Plane(p,vec);
}

FT DistanceField::box_diag()
{
    if (diag < FT(0)) {
        double dx = _m_bbox->xmax()-_m_bbox->xmin();
        double dy = _m_bbox->ymax()-_m_bbox->ymin();
        double dz = _m_bbox->zmax()-_m_bbox->zmin();
        diag = FT(std::sqrt(dx*dx + dy*dy + dz*dz));
    }
    return diag;
}

template <typename Tree, typename Input>
void DistanceField::eval_distance_function_slice(const Tree& tree, Input* input, const Aff_transformation& t, const int m_grid_size)
{
    // Get transformation
    m_max_distance_function = FT(0);
    diag = box_diag();

    const FT dx = diag;
    const FT dy = diag;
    const FT z(0);
    const FT fd =  FT(2);
    for(int i=0 ; i<m_grid_size ; ++i)
    {
        FT x = -diag/fd + FT(i)/FT(m_grid_size) * dx;
        for(int j=0 ; j<m_grid_size ; ++j)
        {
            FT y = -diag/fd + FT(j)/FT(m_grid_size) * dy;
            const Point query = t( Point(x,y,z) );
            FT out_dist(-1.);
            Vector out_grad(0.,0.,0.);
            compute_dist_grad(tree, input, query, out_dist, out_grad);

            m_distance_function[i][j] = Point_distance(query, out_dist);
            m_gradient_function[i][j] = Point_with_normal(query, out_grad);
            m_max_distance_function = (std::max)(out_dist, m_max_distance_function);
        }
    }
}

template <typename Tree>
void DistanceField::sign_distance_function_slice(const Tree& tree, const int m_grid_size)
{
    typedef typename Tree::size_type size_type;
    Vector random_vec = random_vector();

    for(int i=0 ; i<m_grid_size ; ++i)
    {
        for(int j=0 ; j<m_grid_size ; ++j)
        {
            const Point& p = m_distance_function[i][j].first;
            const FT unsigned_distance = m_distance_function[i][j].second;
            const Vector unsigned_gradient = m_gradient_function[i][j].second;
            // get sign through ray casting (random vector)
            Ray ray(p, random_vec);
            size_type nbi = tree.number_of_intersected_primitives(ray);

            FT sign ( (nbi&1) == 0 ? 1 : -1);
            m_distance_function[i][j].second = sign * unsigned_distance;
            m_gradient_function[i][j].second = sign * unsigned_gradient;
        }
    }
}

template <typename Tree, typename Input>
void DistanceField::eval_distance_function(const Tree& tree, Input* input, 
        const std::vector<Point>& queryList, 
        std::vector<Point_distance>& out_dist, 
        std::vector<Point_with_normal>& out_grad)
{
    int nbq = int(queryList.size());
    out_dist.clear();
    out_grad.clear();
    out_dist.reserve(nbq);
    out_grad.reserve(nbq);

    for(int i = 0; i < nbq; ++i)
    {
        const Point& query = queryList[i];
        FT dist(-1.);
        Vector grad(0.,0.,0.);
        compute_dist_grad(tree, input, query, dist, grad);
        
        out_dist.emplace_back(query,dist);
        out_grad.emplace_back(query,grad);
    }
}

template <typename Tree, typename Input>
void DistanceField::eval_distance_function(const Tree& tree, Input* input, 
        const std::vector<Point_with_normal>& queryList, 
        std::vector<Point_distance>& out_dist, 
        std::vector<Point_with_normal>& out_grad)
{
    int nbq = int(queryList.size());
    out_dist.clear();
    out_grad.clear();
    out_dist.reserve(nbq);
    out_grad.reserve(nbq);

    for(int i = 0; i < nbq; ++i)
    {
        const Point& query = queryList[i].first;
        FT dist(-1.);
        Vector grad(0.,0.,0.);
        compute_dist_grad(tree, input, query, dist, grad);
        
        out_dist.emplace_back(query,dist);
        out_grad.emplace_back(query,grad);
    }
}

template <typename Tree>
void DistanceField::sign_distance_function(const Tree& tree, 
        std::vector<Point_distance>& out_dist, 
        std::vector<Point_with_normal>& out_grad)
{
    typedef typename Tree::size_type size_type;
    Vector random_vec = random_vector();
    int nbq = int(out_dist.size());

    for(int i=0 ; i<nbq ; ++i)
    {
        const Point& p = out_dist[i].first;
        const FT unsigned_distance = out_dist[i].second;
        const Vector unsigned_gradient = out_grad[i].second;
        // get sign through ray casting (random vector)
        Ray ray(p, random_vec);
        size_type nbi = tree.number_of_intersected_primitives(ray);

        FT sign ( (nbi&1) == 0 ? 1 : -1); // 1 for outside, -1 for inside
        out_dist[i].second = sign * unsigned_distance;
        out_grad[i].second = sign * unsigned_gradient;
    }
}


void DistanceField::compute_dist_grad(const Facet_tree& tree, Mesh* m_input, const Point& query, FT& out_dist, Vector& out_grad)
{
    typedef typename Facet_tree::Point_and_primitive_id Point_and_primitive_id;
    Point_and_primitive_id pp = tree.closest_point_and_primitive(query);
    Vector displacement = query - pp.first;
    out_dist = CGAL::sqrt( displacement.squared_length() );
    if (out_dist>FT(1e-6)) {
        out_grad = displacement/out_dist;
    } else {
        auto fnormals = m_input->property_map<Face_index, Vector>("f:normal").first;
        out_grad = fnormals[pp.second];
    }
}


void DistanceField::compute_dist_grad(const Triangle_tree& tree, TriangleList* m_input, const Point& query, FT& out_dist, Vector& out_grad)
{
    typedef typename Triangle_tree::Point_and_primitive_id Point_and_primitive_id;
    Point_and_primitive_id pp = tree.closest_point_and_primitive(query);
    Vector displacement = query - pp.first;
    out_dist = CGAL::sqrt( displacement.squared_length() );
    if (out_dist>FT(1e-6)) {
        out_grad = displacement/out_dist;
    } else {
        TriangleList::iterator it = pp.second;
        out_grad = CGAL::normal(it->vertex(0), it->vertex(1), it->vertex(2));
    }
}

template <typename Tree, typename Input>
void DistanceField::compute_distance_function(const Tree& tree, Input* input, const Aff_transformation& t, const int m_grid_size)
{
    
    if ( tree.empty() ) { 
        std::cerr << "AABB tree is empty!" << std::endl;
        return; 
    }
    eval_distance_function_slice(tree, input, t, m_grid_size);
    //sign_distance_function_slice(tree, m_grid_size);
    m_cutting_plane_ready = true;
}

template <typename Tree, typename Input, typename PointType>
void DistanceField::compute_distance_function(const Tree& tree, Input* input, 
        const std::vector<PointType>& queryList, 
        std::vector<Point_distance>& out_dist, 
        std::vector<Point_with_normal>& out_grad) 
{
    
    if ( tree.empty() ) { 
        std::cerr << "AABB tree is empty!" << std::endl;
        return; 
    }
    eval_distance_function(tree, input, queryList, out_dist, out_grad);
    //sign_distance_function(tree, out_dist, out_grad);
}

void MeshDistanceField::compute_distance_function(const Aff_transformation& t, const int m_grid_size) 
{
    // Building the tree if it is not already built
    build_tree();
    Base::compute_distance_function<Facet_tree, Mesh>(m_tree, _m_input, t, m_grid_size);
}

void MeshDistanceField::compute_distance_function(const std::vector<Point_with_normal>& queryList, 
            std::vector<Point_distance>& out_dist, std::vector<Point_with_normal>& out_grad) 
{
    // Building the tree if it is not already built
    build_tree();
    Base::compute_distance_function<Facet_tree, Mesh, Point_with_normal>(m_tree, _m_input, queryList, out_dist, out_grad);
}

void MeshDistanceField::compute_distance_function(const std::vector<Point>& queryList, 
            std::vector<Point_distance>& out_dist, std::vector<Point_with_normal>& out_grad) 
{
    // Building the tree if it is not already built
    build_tree();
    Base::compute_distance_function<Facet_tree, Mesh, Point>(m_tree, _m_input, queryList, out_dist, out_grad);
}

void MeshDistanceField::compute_grad(const Point& query, Vector& out_grad) {
    build_tree();
    FT dist(-1.);
    Base::compute_dist_grad(m_tree, _m_input, query, dist, out_grad);
}

void MeshDistanceField::compute_grad(const std::vector<Point>& queryList, std::vector<Vector>& out_gradList) {
    build_tree();
    out_gradList.clear();
    out_gradList.reserve(queryList.size());
    
    for(size_t i = 0; i < queryList.size(); ++i)
    {
        const Point& query = queryList[i];
        FT dist(-1.);
        Vector grad(0.,0.,0.);
        Base::compute_dist_grad(m_tree, _m_input, query, dist, grad);   
        out_gradList.push_back(grad);
    }
}


void TriangleDistanceField::compute_distance_function(const Aff_transformation& t, const int m_grid_size) 
{
    // Building the tree if it is not already built
    build_tree();
    Base::compute_distance_function<Triangle_tree, TriangleList>(m_tree, _m_input, t, m_grid_size);
}

void TriangleDistanceField::compute_distance_function(const std::vector<Point>& queryList, 
            std::vector<Point_distance>& out_dist, std::vector<Point_with_normal>& out_grad) 
{
    // Building the tree if it is not already built
    build_tree();
    Base::compute_distance_function<Triangle_tree, TriangleList, Point>(m_tree, _m_input, queryList, out_dist, out_grad);
}

void TriangleDistanceField::compute_distance_function(const std::vector<Point_with_normal>& queryList, 
            std::vector<Point_distance>& out_dist, std::vector<Point_with_normal>& out_grad) 
{
    // Building the tree if it is not already built
    build_tree();
    Base::compute_distance_function<Triangle_tree, TriangleList, Point_with_normal>(m_tree, _m_input, queryList, out_dist, out_grad);
}

void TriangleDistanceField::compute_grad(const Point& query, Vector& out_grad) {
    build_tree();
    FT dist(-1.);
    Base::compute_dist_grad(m_tree, _m_input, query, dist, out_grad);
}

void TriangleDistanceField::compute_grad(const std::vector<Point>& queryList, std::vector<Vector>& out_gradList) {
    build_tree();
    out_gradList.clear();
    out_gradList.reserve(queryList.size());
    
    for(size_t i = 0; i < queryList.size(); ++i)
    {
        const Point& query = queryList[i];
        FT dist(-1.);
        Vector grad(0.,0.,0.);
        Base::compute_dist_grad(m_tree, _m_input, query, dist, grad);   
        out_gradList.push_back(grad);
    }
}