#ifndef _CLUSTERING_
#define _CLUSTERING_

#include "types.h"
#include <algorithm>
#include <cstdlib>
#include <limits>
#include <random>
#include <vector>
#include "kabsch.h"
#include "cosine_similarity.h"

using DataFrame = std::vector<Vector>;


std::pair<DataFrame, DataFrame> joint_em_semi_sphere(const DataFrame& data_x, 
                  const DataFrame& data_y, 
                  const size_t k,
                  size_t number_of_iterations, double lambda) {
  // static std::random_device seed;
  // static std::mt19937 random_number_generator(seed());
  static std::mt19937 random_number_generator(0);
  std::uniform_int_distribution<size_t> indices_x(0, data_x.size() - 1);
  std::uniform_int_distribution<size_t> indices_y(0, data_y.size() - 1);
  std::uniform_real_distribution<double> dis(-1.0, 1.0);

  // Pick centroids as random points from the dataset.
  DataFrame means_x(k), means_y(k);
  std::vector<double> stds_x(k, 0.1), stds_y(k, 0.1);

  for (size_t i=0; i<means_x.size(); i++) {
    means_x[i] = data_x[indices_x(random_number_generator)];
    // Ensure the current centroid is different from previous ones
    size_t j=0;
    while (j<i) {
      if ( std::abs(cosine_similarity(means_x[i], means_x[j]))>0.75 || means_x[i].squared_length()<1e-5) {
        Vector perturb(dis(random_number_generator), dis(random_number_generator), dis(random_number_generator));
        means_x[i] = data_x[indices_x(random_number_generator)];
        means_x[i] += perturb;
        means_x[i] /= std::max(std::sqrt(means_x[i].squared_length()), 1e-6);
        j=0;
      } else j++;
    }
  }
  Vector cross_n01 = CGAL::cross_product(means_x[0], means_x[1]);
  if (cross_n01*means_x[2] < 0) means_x[2] *= -1;
  
  // Init R to be an identity, therefore the initial centroids of data_y will be the same
  // R maps x (source) to y (target)
  means_y[0] = Vector(means_x[0]);
  means_y[1] = Vector(means_x[1]);
  means_y[2] = Vector(means_x[2]);
	Eigen::Matrix3d R = Eigen::Matrix3d::Identity(3, 3);
  std::vector<int> order = {0,1,2};

  DataFrame new_means_x(k), new_means_y(k);
  std::vector<double> new_stds_x(k), new_stds_y(k);
  std::vector<std::vector<double>> responsibility_x(data_x.size(), std::vector<double>(k)), responsibility_y(data_y.size(), std::vector<double>(k));
  std::vector<double> pi_x(k, 1.0/k), pi_y(k, 1.0/k);
  for (size_t iteration = 0; iteration < number_of_iterations; ++iteration) {
    // E step
    for (size_t point = 0; point < data_x.size(); ++point) {
      double total_prob = 0.;
      for (size_t cluster = 0; cluster < k; ++cluster) {
        double prob = pi_x[cluster] * calculate_3d_probability_cosine(data_x[point], means_x[cluster], stds_x[cluster]);
        total_prob += prob;
        responsibility_x[point][cluster] = prob;
      }
      if(total_prob > 0) {
          for(size_t cluster = 0; cluster < k; ++cluster) {
              responsibility_x[point][cluster] /= total_prob;
          }
      }
    }
    for (size_t point = 0; point < data_y.size(); ++point) {
      double total_prob = 0.;
      for (size_t cluster = 0; cluster < k; ++cluster) {
        double prob = pi_y[cluster] * calculate_3d_probability_cosine(data_y[point], means_y[cluster], stds_y[cluster]);
        total_prob += prob;
        responsibility_y[point][cluster] = prob;
      }
      if(total_prob > 0) {
          for(size_t cluster = 0; cluster < k; ++cluster) {
              responsibility_y[point][cluster] /= total_prob;
          }
      }
    }

    // M step
    std::vector<double> nc_x(k, 0.), nc_y(k, 0.);
    for (size_t point = 0; point < data_x.size(); ++point) {
      for (size_t cluster = 0; cluster < k; ++cluster) {
        nc_x[cluster] += responsibility_x[point][cluster];
      }
    }
    for (size_t point = 0; point < data_y.size(); ++point) {
      for (size_t cluster = 0; cluster < k; ++cluster) {
        nc_y[cluster] += responsibility_y[point][cluster];
      }
    }
    for (size_t cluster = 0; cluster < k; ++cluster) {
      pi_x[cluster] = nc_x[cluster]/std::max<size_t>(1, data_x.size());
      pi_y[cluster] = nc_y[cluster]/std::max<size_t>(1, data_y.size());
    }

    for (size_t cluster = 0; cluster < k; ++cluster) {
      // compute mean
      new_means_x[cluster]=Vector(0.,0.,0.);
      for (size_t point = 0; point < data_x.size(); ++point) {
        new_means_x[cluster] += data_x[point]*means_x[cluster]>=0 ? responsibility_x[point][cluster] * data_x[point] : -responsibility_x[point][cluster] * data_x[point];
      }
      if (nc_x[cluster]>0) {
        new_means_x[cluster] += 2*lambda*data_x.size()*new_means_y[cluster];
        new_means_x[cluster] /= (nc_x[cluster] + 2*lambda*data_x.size());
      }
      for (Vector& v: new_means_x) v /= std::max(std::sqrt(v.squared_length()), 1e-6);
      new_means_y[cluster]=Vector(0.,0.,0.);
      for (size_t point = 0; point < data_y.size(); ++point) {
        new_means_y[cluster] += data_y[point]*means_y[cluster]>=0 ? responsibility_y[point][cluster] * data_y[point] : -responsibility_y[point][cluster] * data_y[point];
      }
      if (nc_y[cluster]>0) {
        new_means_y[cluster] += 2*lambda*data_y.size()*new_means_x[order[cluster]];
        new_means_y[cluster] /= (nc_y[cluster] + 2*lambda*data_y.size());
      }
      for (Vector& v: new_means_y) v /= std::max(std::sqrt(v.squared_length()), 1e-6);
      // compute R
      Eigen::Matrix3Xd src_set(3, 3), tgt_set(3, 3);
      for (int col = 0; col < src_set.cols(); col++) {
        tgt_set(0, col) = new_means_y[col].x();
        tgt_set(1, col) = new_means_y[col].y();
        tgt_set(2, col) = new_means_y[col].z();
        src_set(0, col) = new_means_x[order[col]].x();
        src_set(1, col) = new_means_x[order[col]].y();
        src_set(2, col) = new_means_x[order[col]].z();
      }
      Eigen::Affine3d A;
      Find3DAffineTransform_unordered(src_set, tgt_set, A, order);
      R = A.rotation();
      // compute std
      new_stds_x[cluster] = 0;
      for (size_t point = 0; point < data_x.size(); ++point) {
        double diff = 1.0 - std::abs(cosine_similarity(data_x[point], new_means_x[cluster]));
        new_stds_x[cluster] += responsibility_x[point][cluster] * diff * diff;
      }
      if (nc_x[cluster]>0) {
        new_stds_x[cluster] /= nc_x[cluster];
      }
      new_stds_x[cluster] = std::max(std::sqrt(new_stds_x[cluster]), 1e-2);

      new_stds_y[cluster] = 0;
      for (size_t point = 0; point < data_y.size(); ++point) {
        double diff = 1.0 - std::abs(cosine_similarity(data_y[point], new_means_y[cluster]));
        new_stds_y[cluster] += responsibility_y[point][cluster] * diff * diff;
      }
      if (nc_y[cluster]>0) {
        new_stds_y[cluster] /= nc_y[cluster];
      }
      new_stds_y[cluster] = std::max(std::sqrt(new_stds_y[cluster]), 1e-2);

      means_x[cluster] = new_means_x[cluster];
      stds_x[cluster] = new_stds_x[cluster];
      means_y[cluster] = new_means_y[cluster];
      stds_y[cluster] = new_stds_y[cluster];
    }
  }

  for (Vector& v: means_x) v /= std::max(std::sqrt(v.squared_length()), 1e-6);
  for (Vector& v: means_y) v /= std::max(std::sqrt(v.squared_length()), 1e-6);
  // reorder means_x according to the computed order
  DataFrame means_x_ordered(k);
  for (int i=0; i<means_x.size(); i++) means_x_ordered[i] = means_x[order[i]];

  return std::pair<DataFrame, DataFrame>(means_x_ordered, means_y);
}


DataFrame em_semi_sphere(const DataFrame& data,
                  const size_t k,
                  size_t number_of_iterations) {
  // static std::random_device seed;
  // static std::mt19937 random_number_generator(seed());
  static std::mt19937 random_number_generator(0);
  std::uniform_int_distribution<size_t> indices(0, data.size() - 1);
  std::uniform_real_distribution<double> dis(-0.2, 0.2);

  // // Pick centroids as random points from the dataset.
  // DataFrame means(k);
  // std::vector<double> stds(k, 0.1);
  // for (size_t i=0; i<means.size(); i++) {
  //   means[i] = data[indices(random_number_generator)];
  //   // Ensure the current centroid is different from previous ones
  //   size_t j=0;
  //   int trial = 0;
  //   while (j<i) {
  //     bool change = false;
  //     if (i==2) { // coplanarity check
  //       Vector temp = CGAL::cross_product(means[0],means[1]);
  //       if (std::abs(temp.x()*means[2].x()+temp.y()*means[2].y()+temp.z()*means[2].z()) < 0.05 ) change=true;
  //     }
  //     if ((!change) && (std::abs(cosine_similarity(means[i], means[j]))>0.75 || means[i].squared_length()<1e-3)) {
  //       change=true;
  //     }
  //     if (change && trial<5000) {
  //       ++trial;
  //       Vector perturb(dis(random_number_generator), dis(random_number_generator), dis(random_number_generator));
  //       means[i] = data[indices(random_number_generator)];
  //       means[i] += perturb;
  //       means[i] /= std::max(std::sqrt(means[i].squared_length()), 1e-6);
  //       j=0;
  //     } else j++;
  //   }
  // }

  // Pick centroids as random points from the dataset.
  DataFrame means(k);
  for (size_t i=0; i<means.size(); i++) {
    if (i==0) means[i] = Vector(1.,0.,0.);
    else if (i==1) means[i] = Vector(0.,1.,0.);
    else if (i==2) means[i] = Vector(0.,0.,1.);
    else {
      means[i] = Vector(dis(random_number_generator), dis(random_number_generator), dis(random_number_generator));
      means[i] /= std::max(std::sqrt(means[i].squared_length()), 1e-6);
      size_t j=0;
      int trial = 0;
      while (j<i) {
        bool change = false;
        if ((!change) && (std::abs(cosine_similarity(means[i], means[j]))>0.75 || means[i].squared_length()<1e-3)) {
          change=true;
        }
        if (change && trial<1000) {
          ++trial;
          means[i] = Vector(dis(random_number_generator), dis(random_number_generator), dis(random_number_generator));
          means[i] /= std::max(std::sqrt(means[i].squared_length()), 1e-6);
          j=0;
        } else j++;
      }
    }
  }
  std::vector<double> stds(k, 0.2);


  DataFrame new_means(k);
  std::vector<double> new_stds(k);
  std::vector<std::vector<double>> responsibility(data.size(), std::vector<double>(k));
  std::vector<double> pi(k, 1.0/k);
  for (size_t iteration = 0; iteration < number_of_iterations; ++iteration) {
    DataFrame old_means = means;
    std::vector<double> old_stds = stds;

    // E step
    for (size_t point = 0; point < data.size(); ++point) {
      double total_prob = 0.;
      for (size_t cluster = 0; cluster < k; ++cluster) {
        double prob = pi[cluster] * calculate_3d_probability_cosine(data[point], means[cluster], stds[cluster]);
        total_prob += prob;
        responsibility[point][cluster] = prob;
      }

      if(total_prob > 0) {
          for(size_t cluster = 0; cluster < k; ++cluster) {
              responsibility[point][cluster] /= total_prob;
          }
      }
    }

    // M step
    std::vector<double> nc(k, 0.);
    for (size_t point = 0; point < data.size(); ++point) {
      for (size_t cluster = 0; cluster < k; ++cluster) {
        nc[cluster] += responsibility[point][cluster];
      }
    }
    for (size_t cluster = 0; cluster < k; ++cluster) {
      pi[cluster] = nc[cluster]/std::max<size_t>(1, data.size());
    }

    for (size_t cluster = 0; cluster < k; ++cluster) {
      // compute mean
      new_means[cluster]=Vector(0.,0.,0.);
      for (size_t point = 0; point < data.size(); ++point) {
        new_means[cluster] += data[point]*means[cluster]>=0 ? responsibility[point][cluster] * data[point] : -responsibility[point][cluster] * data[point];
      }
      if (nc[cluster]>0) {
        new_means[cluster] /= nc[cluster];
      }
      // compute std
      new_stds[cluster] = 0;
      for (size_t point = 0; point < data.size(); ++point) {
        double diff = 1.0 - std::abs(cosine_similarity(data[point], new_means[cluster]));
        new_stds[cluster] += responsibility[point][cluster] * diff * diff;
      }
      if (nc[cluster]>0) {
        new_stds[cluster] /= nc[cluster];
      }
      new_stds[cluster] = std::max(std::sqrt(new_stds[cluster]), 1e-2);

      means[cluster] = new_means[cluster];
      stds[cluster] = new_stds[cluster];
    }
    // stopping criteria
    bool stop = true;
    for (size_t cluster = 0; cluster < k; ++cluster) {
      if (std::abs(cosine_similarity(old_means[cluster], means[cluster]))<0.999999 || std::abs(stds[cluster]-old_stds[cluster])/(old_stds[cluster]+1e-5) > 1e-5 ) {
        stop = false;
      }
    }
    if (stop) break;
  }

  for (Vector& v: means) v /= std::max(std::sqrt(v.squared_length()), 1e-6);
  return means;
}


void em_semi_sphere(const DataFrame& data,
                  const std::vector<double>& weights, 
                  const size_t k,
                  size_t number_of_iterations,
                  DataFrame& means,
                  std::vector<double>& stds) {
  
  double total_weight = 0;
  for (double w: weights) total_weight += w;

  // Pick centroids randomly (with weights)
  means = DataFrame(k, Vector());
  stds = std::vector<double>(k, 0.2);

  static std::mt19937 random_number_generator(0);
  std::discrete_distribution<int> index_dist(std::begin(weights), std::end(weights));
  std::uniform_real_distribution<double> dis(-1.0, 1.0);
  for (size_t i=0; i<means.size(); i++) {
    if (i==0) means[i] = Vector(1.,0.,0.);
    else if (i==1) means[i] = Vector(0.,1.,0.);
    else if (i==2) means[i] = Vector(0.,0.,1.);
    else {
      means[i] = Vector(dis(random_number_generator), dis(random_number_generator), dis(random_number_generator));
      means[i] /= std::max(std::sqrt(means[i].squared_length()), 1e-6);
      size_t j=0;
      int trial = 0;
      while (j<i) {
        bool change = false;
        if ((!change) && (std::abs(cosine_similarity(means[i], means[j]))>0.75 || means[i].squared_length()<1e-3)) {
          change=true;
        }
        if (change && trial<1000) {
          ++trial;
          means[i] = Vector(dis(random_number_generator), dis(random_number_generator), dis(random_number_generator));
          means[i] /= std::max(std::sqrt(means[i].squared_length()), 1e-6);
          j=0;
        } else j++;
      }
    }
  }

  DataFrame new_means(k);
  std::vector<double> new_stds(k);
  std::vector<std::vector<double>> responsibility(data.size(), std::vector<double>(k));
  std::vector<double> pi(k, 1.0/k);
  for (size_t iteration = 0; iteration < number_of_iterations; ++iteration) {
    DataFrame old_means = means;
    std::vector<double> old_stds = stds;

    // E step
    for (size_t point = 0; point < data.size(); ++point) {
      double total_prob = 0.;
      for (size_t cluster = 0; cluster < k; ++cluster) {
        double prob = pi[cluster] * calculate_3d_probability_cosine(data[point], means[cluster], stds[cluster]) * weights[point];
        total_prob += prob;
        responsibility[point][cluster] = prob;
      }

      if(total_prob > 0) {
          for(size_t cluster = 0; cluster < k; ++cluster) {
              responsibility[point][cluster] /= total_prob;
          }
      }
    }

    // M step
    std::vector<double> nc(k, 0.);
    for (size_t point = 0; point < data.size(); ++point) {
      for (size_t cluster = 0; cluster < k; ++cluster) {
        nc[cluster] += responsibility[point][cluster];
      }
    }
    for (size_t cluster = 0; cluster < k; ++cluster) {
      pi[cluster] = nc[cluster]/total_weight;
    }

    for (size_t cluster = 0; cluster < k; ++cluster) {
      // compute mean
      new_means[cluster]=Vector(0.,0.,0.);
      for (size_t point = 0; point < data.size(); ++point) {
        new_means[cluster] += data[point]*means[cluster]>=0 ? responsibility[point][cluster] * data[point] : -responsibility[point][cluster] * data[point];
      }
      if (nc[cluster]>0) {
        new_means[cluster] /= nc[cluster];
      }
      new_means[cluster] /= std::max(std::sqrt(new_means[cluster].squared_length()), 1e-6);
      // compute std
      new_stds[cluster] = 0;
      for (size_t point = 0; point < data.size(); ++point) {
        double diff = 1.0 - std::abs(cosine_similarity(data[point], new_means[cluster]));
        new_stds[cluster] += responsibility[point][cluster] * diff * diff;
      }
      if (nc[cluster]>0) {
        new_stds[cluster] /= nc[cluster];
      }
      new_stds[cluster] = std::max(std::sqrt(new_stds[cluster]), 1e-2);

      means[cluster] = new_means[cluster];
      stds[cluster] = new_stds[cluster];
    }

    // stopping criteria
    bool stop = true;
    for (size_t cluster = 0; cluster < k; ++cluster) {
      if (std::abs(cosine_similarity(old_means[cluster], means[cluster]))<0.999999 || std::abs(stds[cluster]-old_stds[cluster])/(old_stds[cluster]+1e-5) > 1e-5 ) {
        stop = false;
      }
    }
    if (stop) break;
  }
}


DataFrame k_means(const DataFrame& data,
                  const size_t k,
                  size_t number_of_iterations) {
  static std::random_device seed;
  static std::mt19937 random_number_generator(seed());
  std::uniform_int_distribution<size_t> indices(0, data.size() - 1);
  std::uniform_real_distribution<double> dis(-1.0, 1.0);

  // Pick centroids as random points from the dataset.
  DataFrame means(k);
  for (size_t i=0; i<means.size(); i++) {
    means[i] = data[indices(random_number_generator)];
    // Ensure the current centroid is different from previous ones
    size_t j=0;
    while (j<i) {
      if ( std::abs(cosine_similarity(means[i], means[j]))>0.75 || means[i].squared_length()<1e-5) {
        Vector perturb(dis(random_number_generator), dis(random_number_generator), dis(random_number_generator));
        means[i] = data[indices(random_number_generator)];
        means[i] += perturb;
        means[i] /= std::max(std::sqrt(means[i].squared_length()), 1e-6);
        j=0;
      } else j++;
    }
  }

  std::vector<size_t> assignments(data.size());
  for (size_t iteration = 0; iteration < number_of_iterations; ++iteration) {
    // Find assignments.
    for (size_t point = 0; point < data.size(); ++point) {
      double best_distance = std::numeric_limits<double>::max();
      size_t best_cluster = 0;
      for (size_t cluster = 0; cluster < k; ++cluster) {
        const double distance =
            squared_l2_distance(data[point], means[cluster]);
        if (distance < best_distance) {
          best_distance = distance;
          best_cluster = cluster;
        }
      }
      assignments[point] = best_cluster;
    }

    // Sum up and count points for each cluster.
    DataFrame new_means(k);
    std::vector<size_t> counts(k, 0);
    for (size_t point = 0; point < data.size(); ++point) {
      const auto cluster = assignments[point];
      new_means[cluster] += data[point];
      counts[cluster] += 1;
    }

    // Divide sums by counts to get new centroids.
    for (size_t cluster = 0; cluster < k; ++cluster) {
      // Turn 0/0 into 0/1 to avoid zero division.
      const auto count = std::max<size_t>(1, counts[cluster]);
      means[cluster] = new_means[cluster] / count;
    }
  }
  return means;
}

DataFrame k_means_semi_sphere(const DataFrame& data,
                  const size_t k,
                  size_t number_of_iterations) {
  // static std::random_device seed;
  // static std::mt19937 random_number_generator(seed());
  static std::mt19937 random_number_generator(0);
  std::uniform_int_distribution<size_t> indices(0, data.size() - 1);

  // Pick centroids as random points from the dataset.
  DataFrame means(k);
  for (auto& cluster : means) {
    cluster = data[indices(random_number_generator)];
  }

  std::vector<size_t> assignments(data.size(), -1);
  for (size_t iteration = 0; iteration < number_of_iterations; ++iteration) {
    // Find assignments.
    size_t change_counter(0);
    for (size_t point = 0; point < data.size(); ++point) {
      double best_distance = std::numeric_limits<double>::max();
      size_t best_cluster = 0;
      for (size_t cluster = 0; cluster < k; ++cluster) {
        const double distance = 
            1.0 - std::abs(cosine_similarity(data[point], means[cluster]));
        if (distance < best_distance) {
          best_distance = distance;
          best_cluster = cluster;
        }
      }
      if (assignments[point]!=best_cluster) change_counter++;
      assignments[point] = best_cluster;
    }

    // Sum up and count points for each cluster.
    DataFrame new_means(k);
    std::vector<size_t> counts(k, 0);
    for (size_t point = 0; point < data.size(); ++point) {
      const auto cluster = assignments[point];
      new_means[cluster] += data[point]*means[cluster]>=0 ? data[point] : -data[point];
      counts[cluster] += 1;
    }

    // Divide sums by counts to get new centroids.
    for (size_t cluster = 0; cluster < k; ++cluster) {
      // Turn 0/0 into 0/1 to avoid zero division.
      const auto count = std::max<size_t>(1, counts[cluster]);
      means[cluster] = new_means[cluster] / count;
    }

    if (change_counter==0) break;
  }
  for (Vector& v: means) v /= std::sqrt(v.squared_length()) + 1e-8;
  return means;
}

#endif