#ifndef _SIGNED_DISTANCE_FIELD_
#define _SIGNED_DISTANCE_FIELD_


#include <CGAL/function.h>
#include <CGAL/AABB_tree.h>
#include <CGAL/AABB_traits.h>
#include <CGAL/AABB_segment_primitive.h>
#include <CGAL/AABB_halfedge_graph_segment_primitive.h>
#include <CGAL/AABB_triangle_primitive.h>
#include <CGAL/AABB_face_graph_triangle_primitive.h>
#include <CGAL/Scale_space_surface_reconstruction_3.h>
#include "types.h"


typedef CGAL::AABB_face_graph_triangle_primitive<Mesh>         Facet_Primitive;
typedef CGAL::AABB_traits<Kernel, Facet_Primitive>             AABB_Facet_Traits;
typedef CGAL::AABB_tree<AABB_Facet_Traits>                     Facet_tree;                

typedef SegmentList::iterator                                  Segment_Iterator;
typedef CGAL::AABB_segment_primitive<Kernel, Segment_Iterator> Segment_Primitive;
typedef CGAL::AABB_traits<Kernel, Segment_Primitive>           AABB_Segment_Traits;
typedef CGAL::AABB_tree<AABB_Segment_Traits>                   Segment_tree;

typedef TriangleList::iterator                                   Triangle_Iterator;
typedef CGAL::AABB_triangle_primitive<Kernel, Triangle_Iterator> Triangle_Primitive;
typedef CGAL::AABB_traits<Kernel, Triangle_Primitive>            AABB_triangle_traits;
typedef CGAL::AABB_tree<AABB_triangle_traits>                    Triangle_tree;


class DistanceField
{
public:
    DistanceField(Bbox *bbox);
	virtual ~DistanceField();
    // clear data
    void clear();

protected:
	// types
    typedef Point3      Point;
    typedef boost::graph_traits<Mesh>::vertex_descriptor         Vertex_Descriptor;
    typedef boost::graph_traits<Mesh>::face_descriptor           Face_Descriptor;
protected:
    Bbox *_m_bbox;
    FT diag;
	// distance functions (simple 2D arrays on cutting plane)
	FT m_max_distance_function;
    Point_distance m_distance_function[100][100];
    Point_with_normal m_gradient_function[100][100];
    bool m_cutting_plane_ready;

protected:
    // utility functions
    Vector random_vector();
    Ray random_ray(const Bbox& bbox);
    Line random_line(const Bbox& bbox);
    Point random_point(const Bbox& bbox);
    Plane random_plane(const Bbox& bbox);
    Segment random_segment(const Bbox& bbox);
    FT random_in(const double a,const double b);
	FT box_diag();

    virtual void build_tree() {}
    
    void compute_dist_grad(const Facet_tree& tree, Mesh* m_input, const Point& query, FT& out_dist, Vector& out_grad);
    void compute_dist_grad(const Triangle_tree& tree, TriangleList* m_input, const Point& query, FT& out_dist, Vector& out_grad);

    template <typename Tree, typename Input>
    void eval_distance_function_slice(const Tree& tree, Input* input, 
            const Aff_transformation& t, const int m_grid_size);
    
    template <typename Tree>
    void sign_distance_function_slice(const Tree& , const int m_grid_size);

    template <typename Tree, typename Input>
    void eval_distance_function(const Tree& tree, Input* input, 
            const std::vector<Point>& queryList, 
            std::vector<Point_distance>& out_dist, 
            std::vector<Point_with_normal>& out_grad);

    template <typename Tree, typename Input>
    void eval_distance_function(const Tree& tree, Input* input, 
            const std::vector<Point_with_normal>& queryList, 
            std::vector<Point_distance>& out_dist, 
            std::vector<Point_with_normal>& out_grad);

    template <typename Tree>
    void sign_distance_function(const Tree& tree, 
        std::vector<Point_distance>& out_dist, 
        std::vector<Point_with_normal>& out_grad);

public:
	// access data
    const FT &max_distance_function() const { return m_max_distance_function; }
    const auto &distance_function() const { return m_distance_function; }
    const auto &gradient_function() const { return m_gradient_function; }
    const bool cutting_plane_ready() { return m_cutting_plane_ready; }

    // distance functions 
    template <typename Tree, typename Input>
    void compute_distance_function(const Tree& tree, Input* input, 
            const Aff_transformation& t, 
            const int m_grid_size); // on a cutting plane

    template <typename Tree, typename Input, typename PointType>
    void compute_distance_function(const Tree& tree, Input* input, 
            const std::vector<PointType>& queryList, 
            std::vector<Point_distance>& out_dist, 
            std::vector<Point_with_normal>& out_grad);
};


class MeshDistanceField : public DistanceField
{
public:
    MeshDistanceField(Bbox *bbox, Mesh *input) 
    : DistanceField(bbox)
    , _m_input(input)
    {
        build_tree();
    }
	~MeshDistanceField() {}
    void clear() {
        m_tree.clear();
        m_cutting_plane_ready = false;
        m_max_distance_function = FT(0);
    }
private:
    typedef DistanceField Base;
	// data
	Mesh *_m_input;
    Facet_tree m_tree;
public:

    void build_tree(){
        if ( NULL == _m_input )
        {
            std::cerr << "Build tree failed: load mesh first." << std::endl;
            return;
        }
        // Don't rebuild tree if it is already built
        if ( !m_tree.empty() ) { return; }
        // build tree
        std::cout << "Building facet tree... ";
        m_tree.rebuild(faces(*_m_input).first, faces(*_m_input).second,*_m_input);
        std::cout<< "Accelerate distance queries... " << std::endl;
        m_tree.accelerate_distance_queries();
        std::cout<< "done" << std::endl;
    }
    void compute_distance_function(const Aff_transformation& t, const int m_grid_size);

    void compute_distance_function(const std::vector<Point>& queryList, 
            std::vector<Point_distance>& out_dist, std::vector<Point_with_normal>& out_grad);

    void compute_distance_function(const std::vector<Point_with_normal>& queryList, 
            std::vector<Point_distance>& out_dist, std::vector<Point_with_normal>& out_grad);

    void compute_grad(const Point& query, Vector& out_grad);
    void compute_grad(const std::vector<Point>& queryList, std::vector<Vector>& out_gradList);
};


class TriangleDistanceField : public DistanceField
{
public:
    TriangleDistanceField(Bbox *bbox, TriangleList *input)
    : DistanceField(bbox)
    , _m_input(input)
    {
        build_tree();
    }
	~TriangleDistanceField() {}
    void clear() {
        m_tree.clear();
        m_cutting_plane_ready = false;
        m_max_distance_function = FT(0);
    }
private:
    typedef DistanceField Base;
	// data
	TriangleList *_m_input;
    Triangle_tree m_tree;
public:

    void build_tree(){
        if ( NULL == _m_input )
        {
            std::cerr << "Build tree failed: load triangles first." << std::endl;
            return;
        }
        // Don't rebuild tree if it is already built
        if ( !m_tree.empty() ) { return; }
        // build tree
        std::cout << "Building triangle tree... ";
        m_tree.rebuild(_m_input->begin(),_m_input->end());
        std::cout<< "Accelerate distance queries... " << std::endl;
        m_tree.accelerate_distance_queries();
        std::cout<< "done" << std::endl;
    }
    void compute_distance_function(const Aff_transformation& t, const int m_grid_size);
    void compute_distance_function(const std::vector<Point>& queryList, 
            std::vector<Point_distance>& out_dist, std::vector<Point_with_normal>& out_grad);
    void compute_distance_function(const std::vector<Point_with_normal>& queryList, 
            std::vector<Point_distance>& out_dist, std::vector<Point_with_normal>& out_grad);
    void compute_grad(const Point& query, Vector& out_grad);
    void compute_grad(const std::vector<Point>& queryList, std::vector<Vector>& out_gradList);
};

#endif // _SIGNED_DISTANCE_FIELD_

