#ifndef _KABSCH_
#define _KABSCH_
// This code is released in public domain

#include <Eigen/Geometry>

// Given two sets of 3D points, find the rotation (+ translation + scale)
// which best maps the first set to the second.
// Source: http://en.wikipedia.org/wiki/Kabsch_algorithm

// The input 3D points are stored as columns.
Eigen::Affine3d Find3DAffineTransform(const Eigen::Matrix3Xd& in, const Eigen::Matrix3Xd& out) {

  // Default output
  Eigen::Affine3d A;
  A.linear() = Eigen::Matrix3d::Identity(3, 3);
  A.translation() = Eigen::Vector3d::Zero();

  if (in.cols() != out.cols())
    throw "Find3DAffineTransform(): input data mis-match";

  // // First find the scale, by finding the ratio of sums of some distances,
  // // then bring the datasets to the same scale.
  // double dist_in = 0, dist_out = 0;
  // for (int col = 0; col < in.cols()-1; col++) {
  //   dist_in  += (in.col(col+1) - in.col(col)).norm();
  //   dist_out += (out.col(col+1) - out.col(col)).norm();
  // }
  // if (dist_in <= 0 || dist_out <= 0)
  //   return A;
  // double scale = dist_out/dist_in;
  // out /= scale;
  double scale = 1.0;

  Eigen::Vector3d in_ctr = Eigen::Vector3d::Zero();
  Eigen::Vector3d out_ctr = Eigen::Vector3d::Zero();
  // // Find the centroids then shift to the origin
  // for (int col = 0; col < in.cols(); col++) {
  //   in_ctr  += in.col(col);
  //   out_ctr += out.col(col);
  // }
  // in_ctr /= in.cols();
  // out_ctr /= out.cols();
  // for (int col = 0; col < in.cols(); col++) {
  //   in.col(col)  -= in_ctr;
  //   out.col(col) -= out_ctr;
  // }

  // SVD
  Eigen::MatrixXd Cov = in * out.transpose();
  Eigen::JacobiSVD<Eigen::MatrixXd> svd(Cov, Eigen::ComputeThinU | Eigen::ComputeThinV);

  // Find the rotation
  double d = (svd.matrixV() * svd.matrixU().transpose()).determinant() > 0 ? 1.0 : -1.0;
  Eigen::Matrix3d I = Eigen::Matrix3d::Identity(3, 3);
  I(2, 2) = d;
  Eigen::Matrix3d R = svd.matrixV() * I * svd.matrixU().transpose();

  // The final transform
  A.linear() = scale * R;
  A.translation() = scale*(out_ctr - R*in_ctr);
  return A;
}

// Given two sets of 3D points, find the rotation
// which best maps the first set to the second.
// The input 3D points are stored as columns.
void Find3DAffineTransform_unordered(const Eigen::Matrix3Xd& in, const Eigen::Matrix3Xd& out, Eigen::Affine3d& A, std::vector<int>& order) {

  // Default output
  A.linear() = Eigen::Matrix3d::Identity(3, 3);
  A.translation() = Eigen::Vector3d::Zero();

	/* Align with all possible combinations, assuming point set has self-consistent normals */
	std::vector<std::vector<int>> combinations = {{0,1,2}, {0,2,1}, {1,0,2}, {1,2,0}, {2,0,1}, {2,1,0}};
  double dist_min = FLT_MAX;
  Eigen::Matrix3Xd src_set(3, 3);
  for (std::vector<int>& ids : combinations) {
    for (int col = 0; col < src_set.cols(); col++) {
      src_set(0, col) = in(0, ids[col]);
      src_set(1, col) = in(1, ids[col]);
      src_set(2, col) = in(2, ids[col]);
    }
    Eigen::Affine3d T = Find3DAffineTransform(src_set, out);
    double dist = (T.rotation()*src_set - out).norm();
    if (dist < dist_min) {
      dist_min = dist;
      A = T;
      order = ids;
    }
  }
}

#endif