#ifndef _REGISTRATION_
#define _REGISTRATION_

#include "distance_field.h"
#include <Eigen/Core>
#include <Eigen/Geometry>
#include "sophus/so3.hpp"
#include "sophus/se3.hpp"
#include "sophus/sim3.hpp"
#include "ceres/ceres.h"
#include <CGAL/Polygon_mesh_processing/distance.h>


class Registration 
{
public:
    Registration(){}
	~Registration(){}

private:
    typedef Point3                    Point;
    typedef Eigen::Matrix<double,6,1> Vector6d;
    typedef Eigen::Matrix<double,7,1> Vector7d;
    typedef Sophus::SE3<double>       SE3;
    typedef Sophus::RxSO3<double>     RxSO3;
    typedef Sophus::Sim3<double>      Sim3;

public:

    static std::vector<Eigen::Matrix3d> init_rotation(const std::vector< std::vector<Point3> >& source, const std::vector<Vector>& source_normal, 
                                                        const std::vector< std::vector<Point3> >& target, const std::vector<Vector>& target_normal, 
                                                        std::vector<Vector>& source_normal_means, std::vector<Vector>& target_normal_means,
                                                        std::vector<double>& source_normal_stds, std::vector<double>& target_normal_stds);
    static std::vector<Eigen::Matrix3d> init_rotation(Pwn_vector& points, Mesh& mesh, MeshDistanceField& sdf_model, std::vector<Vector>& normal_means, std::vector<Vector>& mesh_normal_means);

    static std::pair<double, double> compute_Rt(const Pwn_vector& points, 
            TriangleDistanceField& sdf_points,
            const Pwn_vector& points_target, 
            TriangleDistanceField& sdf_points_target,
            Eigen::Matrix3d& R, 
            Eigen::Vector3d& t,
            double& scale,
            bool scale_est);

    static std::pair<double, double> compute_Rt(const Pwn_vector& points,  
            TriangleDistanceField& sdf_points,
            Mesh& mesh, 
            MeshDistanceField& sdf_model, 
            Eigen::Matrix3d& R, 
            Eigen::Vector3d& t,
            double& scale,
            bool scale_est);

    template <typename Transformation>
    static void transform(const Transformation& T, std::vector<Point>& points);

    template <typename Transformation>
    static void transform(const Transformation &T, std::vector<Point_with_normal> &points);
};


template <typename LieGroup, typename LieAlgebra>
class PointToPrimitiveEvaluationCallback : public ceres::EvaluationCallback {
public:
    PointToPrimitiveEvaluationCallback(double* parameter)
        : ceres::EvaluationCallback(),
        user_parameter_block(parameter),
        prepare_num_calls(0),
        prepare_requested_jacobians(false),
        prepare_new_evaluation_point(false) {}

    void PrepareForEvaluation(bool evaluate_jacobians, bool new_evaluation_point) final {
        LieAlgebra m_gen = Eigen::Map<const LieAlgebra>(user_parameter_block);
        T = LieGroup::exp(m_gen);
        T_inv = LieGroup::exp(-m_gen);
        prepare_num_calls++;
    }

public:
    double* user_parameter_block;
    int prepare_num_calls;
    bool prepare_requested_jacobians;
    bool prepare_new_evaluation_point;
    LieGroup T;
    LieGroup T_inv;
};


template <typename LieGroup, typename LieAlgebra, typename DF>
class PointToPrimitiveCostFunction : public ceres::CostFunction
{
private:
    typedef Point3                                      Point;
    typedef Eigen::Matrix<double,6,1>                   Vector6d;
    typedef Eigen::Matrix<double,7,1>                   Vector7d;
public:
    PointToPrimitiveCostFunction(const Point_with_normal* point, 
                        DF* sdf, LieGroup* T, LieGroup* T_inv, const double& lambda, bool model_sdf, bool scale_est)
         : point_(point), sdf_(sdf), T_(T), T_inv_(T_inv), lambda_(lambda), is_model_sdf(model_sdf), scale_estimation(scale_est)  {
        if (std::is_same<LieAlgebra, Vector6d>::value) {
            mutable_parameter_block_sizes()->push_back(6);
        } else if (std::is_same<LieAlgebra, Vector7d>::value) {
            mutable_parameter_block_sizes()->push_back(7);
        }
        set_num_residuals(1);
    }
     
    virtual bool Evaluate(double const* const* parameters, 
                double* residuals,
                double** jacobians) const {

        std::vector<Point_with_normal> point_copy = {*point_};
        if (is_model_sdf) Registration::transform(*T_, point_copy);
        else Registration::transform(*T_inv_, point_copy);

        double rescale = is_model_sdf? 1./ T_->scale(): 1.;

        const std::vector<Point_with_normal>& const_queryList = point_copy;
        std::vector<Point_distance> out_dist;
        std::vector<Point_with_normal> out_grad;
        sdf_->compute_distance_function(const_queryList, out_dist, out_grad);
        residuals[0] = rescale * out_dist[0].second;

        if (jacobians != NULL &&  jacobians[0] != NULL) {
            const Point_with_normal& p = const_queryList[0];
            const Vector& v = is_model_sdf? rescale*out_grad[0].second
                                    : -rescale*out_grad[0].second;
            jacobians[0][0] = v.x();
            jacobians[0][1] = v.y();
            jacobians[0][2] = v.z();
            jacobians[0][3] = -v.y()*p.first.z() + v.z()*p.first.y();
            jacobians[0][4] = v.x()*p.first.z() - v.z()*p.first.x();
            jacobians[0][5] = -v.x()*p.first.y() + v.y()*p.first.x();
            if (parameter_block_sizes()[0]==7) jacobians[0][6] = scale_estimation? v.x()*p.first.x()+v.y()*p.first.y()+v.z()*p.first.z() : 0.;
        }


        // // second term involving normals
        // if (jacobians[0] != NULL) {
        //     const Point_with_normal& p = const_queryList[0];
        //     const Vector& v = out_grad[0].second;
        //     double sdf_dir = is_model_sdf? 1 : -1;
        //     double a = -lambda_*(v.x()*p.second.x()+v.y()*p.second.y()+v.z()*p.second.z());
        //     std::vector<Point> p_d = {Point(p.first.x()+1e-3,p.first.y(),p.first.z()), Point(p.first.x(),p.first.y()+1e-3,p.first.z()), Point(p.first.x(),p.first.y(),p.first.z()+1e-3)};
        //     std::vector<Vector> grad_d;
        //     sdf_->compute_grad(p_d, grad_d);
        //     Vector dgrad_v = 1e3*((grad_d[0]-v)*v.x() + (grad_d[1]-v)*v.y() + (grad_d[2]-v)*v.z());

        //     jacobians[0][0] += sdf_dir*a*dgrad_v.x();
        //     jacobians[0][1] += sdf_dir*a*dgrad_v.y();
        //     jacobians[0][2] += sdf_dir*a*dgrad_v.z();
        //     jacobians[0][3] += sdf_dir*a*(-dgrad_v.y()*p.first.z() + dgrad_v.z()*p.first.y()  -v.y()*p.second.z() + v.z()*p.second.y());
        //     jacobians[0][4] += sdf_dir*a*(dgrad_v.x()*p.first.z() - dgrad_v.z()*p.first.x()  +v.x()*p.second.z() - v.z()*p.second.x());
        //     jacobians[0][5] += sdf_dir*a*(-dgrad_v.x()*p.first.y() + dgrad_v.y()*p.first.x()  -v.x()*p.second.y() + v.y()*p.second.x());
        //     if (m_gen.size()==7) {
        //         jacobians[0][6] += sdf_dir*a*(dgrad_v.x()*p.first.x()+dgrad_v.y()*p.first.y()+dgrad_v.z()*p.first.z());
        //     }
        // }
        return true;
    }

    static ceres::CostFunction* Create(const Point_with_normal* points, 
                        DF* sdf, LieGroup* T, LieGroup* T_inv, double& lambda, bool model_sdf, bool scale_est=true)
    {
        return new PointToPrimitiveCostFunction(points, sdf, T, T_inv, lambda, model_sdf, scale_est);
    }

private:
    const Point_with_normal* point_;
    DF* sdf_;
    LieGroup* const T_;
    LieGroup* const T_inv_;
    bool is_model_sdf;
    double lambda_;
    bool scale_estimation;
};


#endif // _REGISTRATION_