#include "primitive_detection.h"
#include "RegionGrowing/cloud_segmentation.h"
#include "RegionGrowing/Visualization_Tools.h"
using RegionGrowing::Point_d;
using RegionGrowing::Point_2d;


namespace Shape_detection {

std::shared_ptr<RegionGrowing::Cloud_segmentation> detect_shape(const Pwn_vector& points, 
                const Parameters &params, 
                Pwn_vector& points_filtered, 
                vector<vector<Point3>>& alpha_shapes,
                std::vector<Vector>& normal_alpha_shape,
                double& average_spacing, 
                bool save) {
    
    float epsilon_ratio = static_cast<float>(params.detection_epsilon_ratio);
    float normal_thresh = static_cast<float>(params.detection_normal_threshold);
	int Nmin = params.detection_min_points;
    int KNN = params.detection_knn;
    RegionGrowing::Cloud_segmentation C;

    // prepare input for region growing input
	auto& HPS = C.HPS;
	HPS.clear();
    size_t num_points = points.size();
	HPS.resize(num_points);
	for (size_t i=0; i<num_points; ++i) {
		HPS[i].position = Point_d(points[i].first);
		HPS[i].normal = RegionGrowing::Vector_3(points[i].second);
	}

    C.set_extrema();

	//Compute K-nearest neighbors 
	C.Compute_Knearest_neighbors(KNN);

    //Region Growing
	C.region_growing(epsilon_ratio, normal_thresh, Nmin);

    // get alpha shapes
    compute_alpha_shapes(C, alpha_shapes, normal_alpha_shape);
    average_spacing = C.average_k_spacing;

    // filter points with primitive assignment
    points_filtered.clear();
    points_filtered.reserve(points.size());
    for(int i=0; i<(int)C.plane_point_index.size();i++){
        for(int k=0; k<(int)C.plane_point_index[i].size(); k++){
            int yy=C.plane_point_index[i][k];
            Point_d pt=C.HPS[yy].position;
            Vector n=C.HPS[yy].normal;
            points_filtered.emplace_back(pt, n);
        }
    }

    // optional save
    boost::filesystem::path path_processed_pcd(params.path_input_pcd);
    std::string basename = path_processed_pcd.stem().string();
    std::string save_dir = params.path_output_dir + separator + basename; 
    if (save) save_envelops_with_normal(C, save_dir);

    return std::make_shared<RegionGrowing::Cloud_segmentation>(C);
}


void compute_alpha_shapes(const RegionGrowing::Cloud_segmentation& C, 
                std::vector<std::vector<Point3>>& alpha_shapes,
                std::vector<Vector>& normal_alpha_shape) {
    
    alpha_shapes.clear();
    normal_alpha_shape.clear();
    int val_lab=0;

    for(int i=0; i<(int)C.plane_point_index.size();i++){
    
        std::vector < Point_d > listp;
        std::vector < Vector > listn;
        for(int k=0; k<(int)C.plane_point_index[i].size(); k++){
            int yy=C.plane_point_index[i][k];
            Point_d pt=C.HPS[yy].position;
            Vector n=C.HPS[yy].normal;
            listp.push_back(pt);
            listn.push_back(n);
        }

        val_lab++;
        RegionGrowing::Plane_3 plane;
        double erro=linear_least_squares_fitting_3(listp.begin(),listp.end(),plane, CGAL::Dimension_tag<0>());

        double normal_dir = 0;
        for (int k=0; k<(int)listn.size(); ++k) {
            normal_dir += listn[k]*plane.orthogonal_vector();
        }
        Vector normal_vector = normal_dir > 0 ? plane.orthogonal_vector() : -plane.orthogonal_vector();

        RegionGrowing::Vector_3 vortho=plane.orthogonal_vector();
        RegionGrowing::Vector_3 b1=plane.base1();
        RegionGrowing::Vector_3 b2=plane.base2();
        FT norme1=sqrt(pow(b1.x(),2)+pow(b1.y(),2)+pow(b1.z(),2)); if(norme1<1e-7){norme1=1e-7;}
        FT norme2=sqrt(pow(b2.x(),2)+pow(b2.y(),2)+pow(b2.z(),2)); if(norme2<1e-7){norme2=1e-7;}

        RegionGrowing::Vector_3 vb1=(1/norme1)*b1;
        RegionGrowing::Vector_3 vb2=(1/norme2)*b2;
        assert(vb1.x()*vb2.x()+vb1.y()*vb2.y()+vb1.z()*vb2.z() < 1e-7);
        assert(vb1.x()*vb1.x()+vb1.y()*vb1.y()+vb1.z()*vb1.z() > 1e-8);
        assert(vb2.x()*vb2.x()+vb2.y()*vb2.y()+vb2.z()*vb2.z() > 1e-8);

        vector< Point_d > poinn;
        vector< Point_2d > poin2;
        poinn.reserve(C.plane_point_index[i].size());
        poin2.reserve(C.plane_point_index[i].size());

        //Point_d pto=plane.projection(Point_d(0.,0.,0.));
        for(int k=0; k<(int)C.plane_point_index[i].size();k++){
            int ind=C.plane_point_index[i][k];
            Point_d pt=C.HPS[ind].position;
            Point_d ptp=plane.projection(pt);
            FT X1=vb1.x()*ptp.x()+vb1.y()*ptp.y()+vb1.z()*ptp.z();
            FT X2=vb2.x()*ptp.x()+vb2.y()*ptp.y()+vb2.z()*ptp.z();
            Point_2d ptp2(X1,X2);
            poinn.push_back(ptp);
            poin2.push_back(ptp2);
        }

        RegionGrowing::Alpha_Shape AlphaShape(poin2.begin(),poin2.end(),FT(std::pow(C.alpha_thresh*C.average_k_spacing,2)),RegionGrowing::Alpha_Shape::REGULARIZED);
        
        std::vector < std::pair < Point_2d, Point_2d> > pairs_edges;
        std::vector < bool > active_pairs;
        for (RegionGrowing::Alpha_Shape::Alpha_shape_edges_iterator it = AlphaShape.alpha_shape_edges_begin(); it != AlphaShape.alpha_shape_edges_end(); ++it){
            RegionGrowing::Alpha_Shape::Edge edg= *it;
            Point_2d pe1=edg.first->vertex(edg.first->ccw(edg.second))->point();
            Point_2d pe2=edg.first->vertex(edg.first->cw(edg.second))->point();
        
            std::pair < Point_2d, Point_2d> pair_ed;
            pair_ed.first=pe1;
            pair_ed.second=pe2;
            pairs_edges.push_back(pair_ed);
            active_pairs.push_back(true);
        }

        std::vector < Point_2d > ranked_pts;
        if(pairs_edges.size()>0){
            ranked_pts.push_back(pairs_edges[0].first); 
            ranked_pts.push_back(pairs_edges[0].second); 
            active_pairs[0]=false;
            Point_2d pt_to_find=pairs_edges[0].second;
            do{
                for(int iter=1;iter<pairs_edges.size();iter++){
                    if(active_pairs[iter]){
                        if(pt_to_find == pairs_edges[iter].first) { ranked_pts.push_back( pairs_edges[iter].second);  pt_to_find=pairs_edges[iter].second; active_pairs[iter]=false; }
                        else if(pt_to_find == pairs_edges[iter].second) { ranked_pts.push_back( pairs_edges[iter].first); pt_to_find=pairs_edges[iter].first; active_pairs[iter]=false; }
                    }
                }
            } while(pt_to_find!=pairs_edges[0].first);
        }

        RegionGrowing::Polygon_2 poly2(ranked_pts.begin(),ranked_pts.end());
        RegionGrowing::CDT2 cdt;

        for(RegionGrowing::Polygon_2::Vertex_iterator it2 = poly2.vertices_begin () ; it2 != poly2.vertices_end ();++it2) cdt.insert(*it2);

        CGAL::make_conforming_Delaunay_2(cdt);

        std::vector < RegionGrowing::Polygon_2 >  partition_polys;
        for ( RegionGrowing::CDT2::Finite_faces_iterator  vit = cdt.finite_faces_begin(); vit != cdt.finite_faces_end(); ++vit){        
                vector<Point_2d> p(3);
                p[0] = vit->vertex (0)->point();
                p[1] = vit->vertex (1)->point();
                p[2] = vit->vertex (2)->point();
                Point_2d bary( (p[0].x()+ p[1].x() + p[2].x())/3.,(p[0].y()+ p[1].y() + p[2].y())/3.);

                if (CGAL::bounded_side_2(poly2.vertices_begin (), poly2.vertices_end (),bary, Kernel()) != CGAL::ON_UNBOUNDED_SIDE)
                {
                    RegionGrowing::Polygon_2 po(p.begin(),p.end());
                    partition_polys.push_back(po);
                    vector<Point_d> pd(3);
                    FT X1=p[0].x()*vb1.x()+p[0].y()*vb2.x()-plane.d()*vortho.x();
                    FT Y1=p[0].x()*vb1.y()+p[0].y()*vb2.y()-plane.d()*vortho.y();
                    FT Z1=p[0].x()*vb1.z()+p[0].y()*vb2.z()-plane.d()*vortho.z();
                    pd[0] = Point_d(X1,Y1,Z1); 
                    FT X2=p[1].x()*vb1.x()+p[1].y()*vb2.x()-plane.d()*vortho.x();
                    FT Y2=p[1].x()*vb1.y()+p[1].y()*vb2.y()-plane.d()*vortho.y();
                    FT Z2=p[1].x()*vb1.z()+p[1].y()*vb2.z()-plane.d()*vortho.z();
                    pd[1] = Point_d(X2,Y2,Z2); 
                    FT X3=p[2].x()*vb1.x()+p[2].y()*vb2.x()-plane.d()*vortho.x();
                    FT Y3=p[2].x()*vb1.y()+p[2].y()*vb2.y()-plane.d()*vortho.y();
                    FT Z3=p[2].x()*vb1.z()+p[2].y()*vb2.z()-plane.d()*vortho.z();
                    pd[2] = Point_d(X3,Y3,Z3);
                    alpha_shapes.push_back(pd);
                    normal_alpha_shape.push_back(normal_vector);
                }
        }
	}
}


void compute_alpha_shapes_anchors(const RegionGrowing::Cloud_segmentation& C, 
                std::vector<Point3>& anchors, 
                double search_radius, 
                std::vector<std::vector<Point3>>& alpha_shapes,
                std::vector<Vector>& normal_alpha_shape) {
    
    alpha_shapes.clear();
    normal_alpha_shape.clear();
    int val_lab=0;

    Tree tree(anchors.begin(), anchors.end());

    for(int i=0; i<(int)C.plane_point_index.size();i++){
    
        std::vector < Point_d > listp;
        std::vector < Vector > listn;
        for(int k=0; k<(int)C.plane_point_index[i].size(); k++){
            int yy=C.plane_point_index[i][k];
            Point_d pt=C.HPS[yy].position;
            Vector n=C.HPS[yy].normal;
            listp.push_back(pt);
            listn.push_back(n);
        }

        val_lab++;
        RegionGrowing::Plane_3 plane;
        double erro=linear_least_squares_fitting_3(listp.begin(),listp.end(),plane, CGAL::Dimension_tag<0>());

        double normal_dir = 0;
        for (int k=0; k<(int)listn.size(); ++k) {
            normal_dir += listn[k]*plane.orthogonal_vector();
        }
        Vector normal_vector = normal_dir > 0 ? plane.orthogonal_vector() : -plane.orthogonal_vector();

        RegionGrowing::Vector_3 vortho=plane.orthogonal_vector();
        RegionGrowing::Vector_3 b1=plane.base1();
        RegionGrowing::Vector_3 b2=plane.base2();
        FT norme1=sqrt(pow(b1.x(),2)+pow(b1.y(),2)+pow(b1.z(),2)); if(norme1<1e-7){norme1=1e-7;}
        FT norme2=sqrt(pow(b2.x(),2)+pow(b2.y(),2)+pow(b2.z(),2)); if(norme2<1e-7){norme2=1e-7;}

        RegionGrowing::Vector_3 vb1=(1/norme1)*b1;
        RegionGrowing::Vector_3 vb2=(1/norme2)*b2;

        vector< Point_d > poinn;
        vector< Point_2d > poin2;
        poinn.reserve(C.plane_point_index[i].size());
        poin2.reserve(C.plane_point_index[i].size());

        for(int k=0; k<(int)C.plane_point_index[i].size();k++){
            int ind=C.plane_point_index[i][k];
            Point_d pt=C.HPS[ind].position;
            Neighbor_search search(tree, pt, 1);
            bool include_point=false;
            for(Neighbor_search::iterator it = search.begin(); it != search.end(); ++it){
                if ( std::sqrt(it->second) <= search_radius ) include_point = true;
            }
            if (include_point) {
                Point_d ptp=plane.projection(pt);
                FT X1=vb1.x()*ptp.x()+vb1.y()*ptp.y()+vb1.z()*ptp.z();
                FT X2=vb2.x()*ptp.x()+vb2.y()*ptp.y()+vb2.z()*ptp.z();
                Point_2d ptp2(X1,X2);
                poinn.push_back(ptp);
                poin2.push_back(ptp2);
            }
        }
        if ((int)poin2.size() < 3) continue;

        RegionGrowing::Alpha_Shape AlphaShape(poin2.begin(),poin2.end(),FT(std::pow(C.alpha_thresh*C.average_k_spacing,2)),RegionGrowing::Alpha_Shape::REGULARIZED);
        
        std::vector < std::pair < Point_2d, Point_2d> > pairs_edges;
        std::vector < bool > active_pairs;
        for (RegionGrowing::Alpha_Shape::Alpha_shape_edges_iterator it = AlphaShape.alpha_shape_edges_begin(); it != AlphaShape.alpha_shape_edges_end(); ++it){
            RegionGrowing::Alpha_Shape::Edge edg= *it;
            Point_2d pe1=edg.first->vertex(edg.first->ccw(edg.second))->point();
            Point_2d pe2=edg.first->vertex(edg.first->cw(edg.second))->point();
        
            std::pair < Point_2d, Point_2d> pair_ed;
            pair_ed.first=pe1;
            pair_ed.second=pe2;
            pairs_edges.push_back(pair_ed);
            active_pairs.push_back(true);
        }

        std::vector < Point_2d > ranked_pts;
        if(pairs_edges.size()>0){
            ranked_pts.push_back(pairs_edges[0].first); 
            ranked_pts.push_back(pairs_edges[0].second); 
            active_pairs[0]=false;
            Point_2d pt_to_find=pairs_edges[0].second;
            do{
                for(int iter=1;iter<pairs_edges.size();iter++){
                    if(active_pairs[iter]){
                        if(pt_to_find == pairs_edges[iter].first) { ranked_pts.push_back( pairs_edges[iter].second);  pt_to_find=pairs_edges[iter].second; active_pairs[iter]=false; }
                        else if(pt_to_find == pairs_edges[iter].second) { ranked_pts.push_back( pairs_edges[iter].first); pt_to_find=pairs_edges[iter].first; active_pairs[iter]=false; }
                    }
                }
            }while(pt_to_find!=pairs_edges[0].first);
        }

        RegionGrowing::Polygon_2 poly2(ranked_pts.begin(),ranked_pts.end());
        RegionGrowing::CDT2 cdt;

        for(RegionGrowing::Polygon_2::Vertex_iterator it2 = poly2.vertices_begin () ; it2 != poly2.vertices_end ();++it2) cdt.insert(*it2);

        CGAL::make_conforming_Delaunay_2(cdt);

        std::vector < RegionGrowing::Polygon_2 >  partition_polys;
        for ( RegionGrowing::CDT2::Finite_faces_iterator  vit = cdt.finite_faces_begin(); vit != cdt.finite_faces_end(); ++vit){        
                vector<Point_2d> p(3);
                p[0] = vit->vertex (0)->point();
                p[1] = vit->vertex (1)->point();
                p[2] = vit->vertex (2)->point();
                Point_2d bary( (p[0].x()+ p[1].x() + p[2].x())/3.,(p[0].y()+ p[1].y() + p[2].y())/3.);

                if (CGAL::bounded_side_2(poly2.vertices_begin (), poly2.vertices_end (),bary, Kernel()) != CGAL::ON_UNBOUNDED_SIDE)
                {
                    RegionGrowing::Polygon_2 po(p.begin(),p.end());
                    partition_polys.push_back(po);
                    vector<Point_d> pd(3);
                    FT X1=p[0].x()*vb1.x()+p[0].y()*vb2.x()-plane.d()*vortho.x();
                    FT Y1=p[0].x()*vb1.y()+p[0].y()*vb2.y()-plane.d()*vortho.y();
                    FT Z1=p[0].x()*vb1.z()+p[0].y()*vb2.z()-plane.d()*vortho.z();
                    pd[0] = Point_d(X1,Y1,Z1); 
                    FT X2=p[1].x()*vb1.x()+p[1].y()*vb2.x()-plane.d()*vortho.x();
                    FT Y2=p[1].x()*vb1.y()+p[1].y()*vb2.y()-plane.d()*vortho.y();
                    FT Z2=p[1].x()*vb1.z()+p[1].y()*vb2.z()-plane.d()*vortho.z();
                    pd[1] = Point_d(X2,Y2,Z2); 
                    FT X3=p[2].x()*vb1.x()+p[2].y()*vb2.x()-plane.d()*vortho.x();
                    FT Y3=p[2].x()*vb1.y()+p[2].y()*vb2.y()-plane.d()*vortho.y();
                    FT Z3=p[2].x()*vb1.z()+p[2].y()*vb2.z()-plane.d()*vortho.z();
                    pd[2] = Point_d(X3,Y3,Z3); 
                    alpha_shapes.push_back(pd);
                    normal_alpha_shape.push_back(normal_vector);
                }
        }

	}
}


}