#include "registration.h"
#include "clustering.h"
#include "kabsch.h"

template <typename Transformation>
void Registration::transform(const Transformation &T, std::vector<Point> &points)
{
    Eigen::Matrix<double, 3, 4> T_mat = T.matrix3x4();
    for (Point &p : points) {
        p = Point(T_mat(0,0)*p.x()+T_mat(0,1)*p.y()+T_mat(0,2)*p.z()+T_mat(0,3),
                  T_mat(1,0)*p.x()+T_mat(1,1)*p.y()+T_mat(1,2)*p.z()+T_mat(1,3),
                  T_mat(2,0)*p.x()+T_mat(2,1)*p.y()+T_mat(2,2)*p.z()+T_mat(2,3));
    }
}

template <typename Transformation>
void Registration::transform(const Transformation &T, std::vector<Point_with_normal> &points)
{
    Eigen::Matrix<double, 3, 3> R_mat = T.rotationMatrix();
    Eigen::Matrix<double, 3, 4> T_mat = T.matrix3x4();
    for (Point_with_normal &p : points)
    {
        p = Point_with_normal(Point(T_mat(0,0)*p.first.x()+T_mat(0,1)*p.first.y()+T_mat(0,2)*p.first.z()+T_mat(0,3),
                                    T_mat(1,0)*p.first.x()+T_mat(1,1)*p.first.y()+T_mat(1,2)*p.first.z()+T_mat(1,3),
                                    T_mat(2,0)*p.first.x()+T_mat(2,1)*p.first.y()+T_mat(2,2)*p.first.z()+T_mat(2,3)), 
                              Vector(R_mat(0,0)*p.second.x()+R_mat(0,1)*p.second.y()+R_mat(0,2)*p.second.z(),
                                     R_mat(1,0)*p.second.x()+R_mat(1,1)*p.second.y()+R_mat(1,2)*p.second.z(),
                                     R_mat(2,0)*p.second.x()+R_mat(2,1)*p.second.y()+R_mat(2,2)*p.second.z()) );
    }
}


std::vector<Eigen::Matrix3d> Registration::init_rotation(const std::vector< std::vector<Point3> >& source, const std::vector<Vector>& source_normal, 
                                                        const std::vector< std::vector<Point3> >& target, const std::vector<Vector>& target_normal, 
                                                        std::vector<Vector>& source_normal_means, std::vector<Vector>& target_normal_means,
                                                        std::vector<double>& source_normal_stds, std::vector<double>& target_normal_stds)
{	
    std::vector<Eigen::Matrix3d> Rlist;
    /* Compute triangle area as weights */
    std::vector<double> source_weights(source.size(), 0), target_weights(target.size(), 0);
    for (size_t i=0; i<source.size(); i++) {
        auto& triangle = source[i];
        source_weights[i] = Kernel::Compute_area_3()(triangle[0],triangle[1],triangle[2]);
    }
    for (size_t i=0; i<target.size(); i++) {
        auto& triangle = target[i];
        target_weights[i] = Kernel::Compute_area_3()(triangle[0],triangle[1],triangle[2]);
    }
	/* Clustering into 3-frames */
    em_semi_sphere(source_normal, source_weights, 3, 10000, source_normal_means, source_normal_stds);
    em_semi_sphere(target_normal, target_weights, 3, 10000, target_normal_means, target_normal_stds);

    /* Align with all possible combinations, assuming point set has self-consistent normals */
	Eigen::Matrix3Xd src_set(3, 3), tgt_set(3, 3);
	for (int col = 0; col < src_set.cols(); col++) {
		tgt_set(0, col) = target_normal_means[col].x();
		tgt_set(1, col) = target_normal_means[col].y();
		tgt_set(2, col) = target_normal_means[col].z();
	}
	Vector cross_n01 = CGAL::cross_product(target_normal_means[0], target_normal_means[1]);
	int is_xyz = cross_n01*target_normal_means[2] > 0 ? 1 : -1;
	std::vector<std::vector<int>> uv_orientations = {{1,1}, {-1,1}, {1,-1}, {-1,-1}};
	std::vector<std::vector<int>> combinations = {{0,1,2}, {0,2,1}, {1,0,2}, {1,2,0}, {2,0,1}, {2,1,0}};
	Rlist.reserve(uv_orientations.size()*combinations.size());
	for (std::vector<int>& uv_ori : uv_orientations) {
		for (std::vector<int>& ids : combinations) {
			Vector cross_n = CGAL::cross_product(uv_ori[0]*source_normal_means[ids[0]], uv_ori[1]*source_normal_means[ids[1]]);
			int z_ori = (cross_n*source_normal_means[ids[2]] > 0)==(is_xyz>0) ?  1 : -1;
			for (int col = 0; col < src_set.cols(); col++) {
				if (col<2) {
					src_set(0, col) = uv_ori[col] * source_normal_means[ids[col]].x();
					src_set(1, col) = uv_ori[col] * source_normal_means[ids[col]].y();
					src_set(2, col) = uv_ori[col] * source_normal_means[ids[col]].z();
				} else {
					src_set(0, col) = z_ori * source_normal_means[ids[col]].x();
					src_set(1, col) = z_ori * source_normal_means[ids[col]].y();
					src_set(2, col) = z_ori * source_normal_means[ids[col]].z();
				}
			}
			Eigen::Affine3d A = Find3DAffineTransform(src_set, tgt_set);
			Rlist.emplace_back(A.rotation());
		}
	}
	return Rlist;
}


std::vector<Eigen::Matrix3d> Registration::init_rotation(Pwn_vector& points, Mesh& mesh, MeshDistanceField& sdf_model, std::vector<Vector>& normal_means, std::vector<Vector>& mesh_normal_means)
{	
	size_t N_samples = 1e5;
	std::vector<Eigen::Matrix3d> Rlist;

	/* Clustering into 3 clusters */
	// point set
	Pwn_vector sample_points;
	sample_points.reserve(N_samples);
	if (points.size()>N_samples) {
		std::sample(points.begin(), points.end(), std::back_inserter(sample_points),
                N_samples, std::mt19937{std::random_device{}()});
	} 
	else sample_points = points;
	std::vector<Vector> sample_normals;
	sample_normals.reserve(sample_points.size());
	for (Point_with_normal& pwn : sample_points) {
		sample_normals.push_back(pwn.second);
	}
	normal_means = em_semi_sphere(sample_normals, 3, 3000);
	// mesh
	std::vector<Point3> mesh_samples;
    mesh_samples.reserve(N_samples);
	CGAL::Polygon_mesh_processing::sample_triangle_mesh(mesh, 
										std::back_inserter(mesh_samples), 
										CGAL::Polygon_mesh_processing::parameters::use_random_uniform_sampling(true).number_of_points_on_faces(N_samples).do_sample_edges(false).do_sample_vertices(false));
    // normals at sample points
    std::vector<Vector> mesh_samples_normals;
    sdf_model.compute_grad(mesh_samples, mesh_samples_normals);
	mesh_normal_means = em_semi_sphere(mesh_samples_normals, 3, 3000);

	/* Align with all possible combinations, assuming point set has self-consistent normals */
	Eigen::Matrix3Xd src_set(3, 3), tgt_set(3, 3);
	for (int col = 0; col < src_set.cols(); col++) {
		tgt_set(0, col) = mesh_normal_means[col].x();
		tgt_set(1, col) = mesh_normal_means[col].y();
		tgt_set(2, col) = mesh_normal_means[col].z();
	}
	Vector cross_n01 = CGAL::cross_product(mesh_normal_means[0], mesh_normal_means[1]);
	int is_xyz = cross_n01*mesh_normal_means[2] > 0 ? 1 : -1;
	std::vector<std::vector<int>> uv_orientations = {{1,1}, {-1,1}, {1,-1}, {-1,-1}};
	std::vector<std::vector<int>> combinations = {{0,1,2}, {0,2,1}, {1,0,2}, {1,2,0}, {2,0,1}, {2,1,0}};
	Rlist.reserve(uv_orientations.size()*combinations.size());
	for (std::vector<int>& uv_ori : uv_orientations) {
		for (std::vector<int>& ids : combinations) {
			Vector cross_n = CGAL::cross_product(uv_ori[0]*normal_means[ids[0]], uv_ori[1]*normal_means[ids[1]]);
			int z_ori = (cross_n*normal_means[ids[2]] > 0)==(is_xyz>0) ?  1 : -1;
			for (int col = 0; col < src_set.cols(); col++) {
				if (col<2) {
					src_set(0, col) = uv_ori[col] * normal_means[ids[col]].x();
					src_set(1, col) = uv_ori[col] * normal_means[ids[col]].y();
					src_set(2, col) = uv_ori[col] * normal_means[ids[col]].z();
				} else {
					src_set(0, col) = z_ori * normal_means[ids[col]].x();
					src_set(1, col) = z_ori * normal_means[ids[col]].y();
					src_set(2, col) = z_ori * normal_means[ids[col]].z();
				}
			}
			Eigen::Affine3d A = Find3DAffineTransform(src_set, tgt_set);
			Rlist.emplace_back(A.rotation());
		}
	}
	return Rlist;
}


std::pair<double, double> Registration::compute_Rt(const Pwn_vector& points, 
        TriangleDistanceField& sdf_points,
        const Pwn_vector& points_target, 
        TriangleDistanceField& sdf_points_target,
        Eigen::Matrix3d& R, 
        Eigen::Vector3d& t,
        double& scale,
        bool scale_est)
{   
    size_t nbq_source = points.size();
    size_t nbq_target = points_target.size();
    std::vector<Point_with_normal> ps_queryList, ps_target_queryList;
    ps_queryList.reserve(nbq_source);
    ps_target_queryList.reserve(nbq_source);
    for (auto& pt: points) ps_queryList.emplace_back(pt);
    for (auto& pt: points_target) ps_target_queryList.emplace_back(pt);

    RxSO3 m_RxSO3(scale, R);
    Sim3 m_Sim3(m_RxSO3, t);
    Vector7d m_sim3 = m_Sim3.log(); 
    double init_sim3[7] = {m_sim3(0),m_sim3(1),m_sim3(2),m_sim3(3),m_sim3(4),m_sim3(5),m_sim3(6)};

    // problem setup
    std::vector<double*> parameter_blocks(1, init_sim3);
    ceres::Problem::Options problem_options;
    PointToPrimitiveEvaluationCallback<Sim3, Vector7d> callback(parameter_blocks[0]); 
    problem_options.evaluation_callback = &callback;
    ceres::Problem problem(problem_options);
    double lambda = 1.;
    bool is_target_sdf = true;
    //auto lossfunction = new ceres::SoftLOneLoss(0.1);
    for (size_t i = 0; i < nbq_source; ++i) {
        problem.AddResidualBlock(
            PointToPrimitiveCostFunction<Sim3, Vector7d, TriangleDistanceField>::Create(&ps_queryList[i], &sdf_points_target, &callback.T, &callback.T_inv, lambda, is_target_sdf, scale_est),
            NULL,
            parameter_blocks);
    }
    if (scale_est) {
        is_target_sdf = false;
        for (size_t i = 0; i < nbq_target; ++i) {
            problem.AddResidualBlock(
                PointToPrimitiveCostFunction<Sim3, Vector7d, TriangleDistanceField>::Create(&ps_target_queryList[i], &sdf_points, &callback.T, &callback.T_inv, lambda, is_target_sdf, scale_est),
                NULL,
                parameter_blocks);
        }
    }
    
    // run the solver
    ceres::Solver::Options options;
    options.trust_region_strategy_type = ceres::DOGLEG;
    options.dogleg_type = ceres::SUBSPACE_DOGLEG;
    // options.parameter_tolerance = 1e-9;
    // options.gradient_tolerance = 1e-10;
    // options.function_tolerance = 1e-7;
    options.use_inner_iterations = false;
    options.use_nonmonotonic_steps = false;
    options.min_relative_decrease = 1e-3;
    options.max_num_consecutive_invalid_steps = 8;
    options.initial_trust_region_radius = 1e4;
    options.num_threads = 12; // need to define CGAL_HAS_THREADS in preprocessor
    options.max_num_iterations = 100;
    options.logging_type = ceres::SILENT;
    options.minimizer_progress_to_stdout = false;
    ceres::Solver::Summary summary;
    ceres::Solve(options, &problem, &summary);
    // report
    std::cout<<summary.BriefReport() <<std::endl;
    m_sim3 = Eigen::Map<Vector7d>(parameter_blocks[0]);
    m_Sim3 = Sim3::exp(m_sim3);
    R = m_Sim3.rotationMatrix();
    t = m_Sim3.translation();
    scale = m_Sim3.scale();
    return std::pair<double, double>(summary.initial_cost, summary.final_cost);
}


std::pair<double, double> Registration::compute_Rt(const Pwn_vector& points, 
        TriangleDistanceField& sdf_points,
        Mesh& mesh, 
        MeshDistanceField& sdf_mesh, 
        Eigen::Matrix3d& R, 
        Eigen::Vector3d& t,
        double& scale,
        bool scale_est) 
{   
    size_t nbq = points.size();
    std::vector<Point_with_normal> ps_queryList;
    ps_queryList.reserve(nbq);
    for (auto& pt: points) ps_queryList.emplace_back(pt);

    // sample from mesh
    std::vector<Point> mesh_samples;
    mesh_samples.reserve(nbq);
    double mesh_area = CGAL::Polygon_mesh_processing::area(mesh);
	CGAL::Polygon_mesh_processing::sample_triangle_mesh(mesh, 
										std::back_inserter(mesh_samples), 
										CGAL::Polygon_mesh_processing::parameters::use_random_uniform_sampling(true).number_of_points_on_faces(nbq).do_sample_edges(false).do_sample_vertices(false));
    size_t nbs = mesh_samples.size();
    // add normals to sample points
    std::vector<Vector> mesh_samples_normals;
    sdf_mesh.compute_grad(mesh_samples, mesh_samples_normals);
    std::vector<Point_with_normal> mesh_queryList;
    mesh_queryList.reserve(nbs);
    for (size_t i=0; i<nbs; ++i) {
        mesh_queryList.emplace_back(mesh_samples[i], mesh_samples_normals[i]);
    }

    RxSO3 m_RxSO3(scale, R);
    Sim3 m_Sim3(m_RxSO3, t);
    Vector7d m_sim3 = m_Sim3.log(); 
    double init_sim3[7] = {m_sim3(0),m_sim3(1),m_sim3(2),m_sim3(3),m_sim3(4),m_sim3(5),m_sim3(6)};

    // problem setup
    std::vector<double*> parameter_blocks(1, init_sim3);
    ceres::Problem::Options problem_options;
    PointToPrimitiveEvaluationCallback<Sim3, Vector7d> callback(parameter_blocks[0]); 
    problem_options.evaluation_callback = &callback;
    ceres::Problem problem(problem_options);
    double lambda = std::sqrt(mesh_area)*1e-2;
    bool is_target_sdf = true;
    for (size_t i = 0; i < nbq; ++i) {
        problem.AddResidualBlock(
            PointToPrimitiveCostFunction<Sim3, Vector7d, MeshDistanceField>::Create(&ps_queryList[i], &sdf_mesh, &callback.T, &callback.T_inv, lambda, is_target_sdf, scale_est),
            NULL,
            parameter_blocks);
    }
    is_target_sdf = false;
    for (size_t i = 0; i < nbs; ++i) {
        problem.AddResidualBlock(
            PointToPrimitiveCostFunction<Sim3, Vector7d, TriangleDistanceField>::Create(&mesh_queryList[i], &sdf_points, &callback.T, &callback.T_inv, lambda, is_target_sdf, scale_est),
            NULL,
            parameter_blocks);
    }

    // run the solver
    ceres::Solver::Options options;
    options.trust_region_strategy_type = ceres::DOGLEG;
    options.dogleg_type = ceres::SUBSPACE_DOGLEG;
    // options.parameter_tolerance = 1e-9;
    // options.gradient_tolerance = 1e-10;
    // options.function_tolerance = 1e-7;
    options.use_inner_iterations = false;
    options.use_nonmonotonic_steps = false;
    options.min_relative_decrease = 1e-3;
    options.max_num_consecutive_invalid_steps = 6;
    options.initial_trust_region_radius = 1e4;
    options.num_threads = 6; // need to define CGAL_HAS_THREADS in preprocessor
    options.max_num_iterations = 100;
    options.logging_type = ceres::SILENT;
    options.minimizer_progress_to_stdout = false;
    ceres::Solver::Summary summary;
    ceres::Solve(options, &problem, &summary);

    // report
    std::cout<<summary.BriefReport() <<std::endl;
    m_sim3 = Eigen::Map<Vector7d>(parameter_blocks[0]);
    m_Sim3 = Sim3::exp(m_sim3);
    R = m_Sim3.rotationMatrix();
    t = m_Sim3.translation();
    scale = m_Sim3.scale();

    return std::pair<double, double>(summary.initial_cost, summary.final_cost);
}
