#define CGAL_DISABLE_ROUNDING_MATH_CHECK

#include "args.h"
#include <algorithm>
#include <CGAL/IO/read_xyz_points.h>
#include <CGAL/IO/write_xyz_points.h>
#include <CGAL/IO/OFF_reader.h>
#include <CGAL/Polygon_mesh_processing/polygon_soup_to_polygon_mesh.h>
#include <CGAL/Polygon_mesh_processing/orientation.h>
#include <CGAL/Polygon_mesh_processing/orient_polygon_soup.h>
#include <CGAL/Polygon_mesh_processing/orient_polygon_soup_extension.h>
#include <CGAL/Polygon_mesh_processing/repair_polygon_soup.h>
#include <CGAL/Polygon_mesh_processing/triangulate_hole.h>
#include <CGAL/Polygon_mesh_processing/border.h>
#include <CGAL/boost/graph/Euler_operations.h>
#include <CGAL/optimal_bounding_box.h>
#include <CGAL/Polygon_mesh_processing/bbox.h>
#include <CGAL/property_map.h>
#include <CGAL/Polygon_mesh_processing/transform.h>
#include <CGAL/Polygon_mesh_processing/compute_normal.h>
#include <fstream>
#include <sstream>
#include <boost/filesystem.hpp>
#include <time.h>
#include "registration.h"
#include "primitive_detection.h"
#include "RegionGrowing/Visualization_Tools.h"
using namespace boost::filesystem;

Bbox m_points_bbox;
Bbox m_mesh_bbox;
Pwn_vector m_input_points;
Pwn_vector m_input_points_filtered;
Mesh m_input_mesh;
std::vector< std::vector<Point3> > m_detected_alpha_shapes;
std::vector<Vector> m_normal_alpha_shape;
bool normals_ready = false;
std::unique_ptr<MeshDistanceField> sdf;
std::unique_ptr<TriangleDistanceField> sdf_points;


template <typename T>
std::string to_string_with_precision(const T a_value, const int n = 8)
{
    std::ostringstream out;
    out.precision(n);
    out << std::fixed << a_value;
    return out.str();
}


void compute_mesh_normals()
{
    if (m_input_mesh.is_empty() || normals_ready) {return;}
	std::cout << "Compute mesh face normals" << std::endl;
    normals_ready = true;	
    Mesh::Property_map<Face_index, Vector> fnormals = m_input_mesh.add_property_map<Face_index, Vector>("f:normal", CGAL::NULL_VECTOR).first;
	auto vnormal_add = m_input_mesh.add_property_map<Vertex_index, Vector>("v:normal", CGAL::NULL_VECTOR);
	if (vnormal_add.second) {
		Mesh::Property_map<Vertex_index, Vector> vnormals = vnormal_add.first;
		CGAL::Polygon_mesh_processing::compute_normals(m_input_mesh,
			vnormals,
			fnormals
		); 
	} else {
		CGAL::Polygon_mesh_processing::compute_face_normals(m_input_mesh,
			fnormals
		);
	}
	// debug
	int i=0;
	for (auto n : fnormals) {
		if (n.squared_length()==0) {
			std::cout << "Surface normal is zero for " << i << std::endl;
		}
		++i;
	}
}


bool is_small_hole(halfedge_descriptor h, Mesh & mesh,
                   double max_hole_area)
{
  CGAL::Bbox_3 hole_bbox;
  for (halfedge_descriptor hc : CGAL::halfedges_around_face(h, mesh))
  {
    const Point3& p = mesh.point(target(hc, mesh));
    hole_bbox += p.bbox();
	double xdiam = hole_bbox.xmax() - hole_bbox.xmin();
	double ydiam = hole_bbox.ymax() - hole_bbox.ymin();
	double zdiam = hole_bbox.zmax() - hole_bbox.zmin();
    // Exit early, to avoid unnecessary traversal of large holes
    if ( xdiam*ydiam > max_hole_area) return false;
    if ( xdiam*zdiam > max_hole_area) return false;
    if ( ydiam*zdiam > max_hole_area) return false;
  }
  return true;
}


int open(std::string filename) 
{
	const boost::filesystem::path filepath = filename;
	std::string ext = filepath.extension().string();
	std::ifstream ifs(filename);
	
	if(!ifs)
	{
		std::cerr << "Unable to open the file " << filename << std::endl;
		return EXIT_FAILURE;
	}
	if(ext == ".xyz" || ext == ".XYZ")
	{	
		m_input_points.clear();	
		m_input_points_filtered.clear();
		m_detected_alpha_shapes.clear();
		m_normal_alpha_shape.clear();
		if (!CGAL::read_xyz_points(ifs,
      	std::back_inserter(m_input_points),
      	CGAL::parameters::point_map(Point_map()).normal_map(Normal_map()))) 
		{
      		std::cerr << "Error: cannot read file " << filename << std::endl;
      		return EXIT_FAILURE;
  		}
		// normalize point normal
		for (size_t i = 0 ; i < m_input_points.size() ; ++i) {
			const Vector & n = m_input_points[i].second;
			const double &x = n.x(), &y = n.y(), &z = n.z();
			double l = std::max(sqrt(x * x + y * y + z * z) , 1e-8);
			m_input_points[i].second = Vector (x / l, y / l, z / l);
		}
	}
	else if(ext == ".off" || ext == ".OFF") 
	{
		m_input_mesh.clear();
		normals_ready = false;
		std::cout << "Read surface mesh... ";

		std::vector<Point3> points;
		std::vector<std::vector<std::size_t> > polygons;
		if(!CGAL::read_OFF(ifs, points, polygons) || points.empty())
		{
			std::cerr << "Cannot open file " << std::endl;
			return EXIT_FAILURE;
		}
		std::cout << "Repair polygon soup... ";
		CGAL::Polygon_mesh_processing::repair_polygon_soup(points, polygons);
		if (!CGAL::Polygon_mesh_processing::orient_polygon_soup(points, polygons)) std::cout << "Orienting polygons failed." << std::endl;
		std::cout << "Polygon soup to polygon mesh... ";
		CGAL::Polygon_mesh_processing::polygon_soup_to_polygon_mesh(points, polygons, m_input_mesh);
		std::cout << "Remove zero-area faces... ";
		// remove zero-area faces
		std::unordered_set<Face_index> faces_to_remove;
		for (auto f: m_input_mesh.faces()) {
			if ( CGAL::Polygon_mesh_processing::face_area(f, m_input_mesh)==0 ) {
				faces_to_remove.insert(f);
				double max_length = 0;
				halfedge_descriptor max_he = m_input_mesh.halfedge(f);
				for (halfedge_descriptor he : CGAL::halfedges_around_face(m_input_mesh.halfedge(f), m_input_mesh)) {
					double he_length = CGAL::squared_distance( m_input_mesh.point(m_input_mesh.source(he)), m_input_mesh.point(m_input_mesh.target(he)));
					if (max_length<he_length) {
						max_he = he;
						max_length = he_length;
					}
				}
				faces_to_remove.insert(m_input_mesh.face(m_input_mesh.opposite(max_he)));
			}
		}
		for (auto f: faces_to_remove) CGAL::Euler::remove_face(m_input_mesh.halfedge(f), m_input_mesh);
		// fill holes
		CGAL::Bbox_3 mesh_bbox = CGAL::Polygon_mesh_processing::bbox(m_input_mesh);
		double mesh_bbox_max_area = std::max( (mesh_bbox.zmax()-mesh_bbox.zmin())*(mesh_bbox.ymax()-mesh_bbox.ymin()), std::max((mesh_bbox.xmax()-mesh_bbox.xmin())*(mesh_bbox.ymax()-mesh_bbox.ymin()),(mesh_bbox.xmax()-mesh_bbox.xmin())*(mesh_bbox.zmax()-mesh_bbox.zmin())) );
		std::vector<halfedge_descriptor> border_cycles;
		CGAL::Polygon_mesh_processing::extract_boundary_cycles(m_input_mesh, std::back_inserter(border_cycles));
		for(halfedge_descriptor h : border_cycles)
		{
			if (!is_small_hole(h, m_input_mesh, 0.25*mesh_bbox_max_area)) continue;
			std::vector<face_descriptor>  patch_facets;
			std::vector<vertex_descriptor> patch_vertices;
			CGAL::Polygon_mesh_processing::triangulate_refine_and_fair_hole(
						m_input_mesh,
						h,
						std::back_inserter(patch_facets),
						std::back_inserter(patch_vertices),
						CGAL::Polygon_mesh_processing::parameters::vertex_point_map(get(CGAL::vertex_point, m_input_mesh)).
						geom_traits(Kernel()));
			std::cout << "* Number of facets in constructed patch: " << patch_facets.size() << std::endl;
		}
		// remove duplicated points and edges by stitching
		CGAL::Polygon_mesh_processing::stitch_borders(m_input_mesh);

		// if(CGAL::is_closed(m_input_mesh)) CGAL::Polygon_mesh_processing::orient(m_input_mesh);

		if (!CGAL::is_triangle_mesh(m_input_mesh)){
			std::cerr << "Input geometry is not triangulated." << std::endl;
			return EXIT_FAILURE;
		}
		compute_mesh_normals();
	}
	else {
		std::cerr << "Unknow extenstion. Please use .xyz for point set and .off for surface mesh" << std::endl;
		return -1;		
	}
	std::cout << "done. (" << m_input_points.size() << " points, " 
						   << m_input_mesh.vertices().size() << " vertices)" << std::endl;
	ifs.close();
}


void transform(const Eigen::Matrix3d& R_, const Eigen::Vector3d& t_, const double& scale_, Pwn_vector& points) 
{
    const Aff_transformation T(R_(0, 0)*scale_, R_(0, 1)*scale_, R_(0, 2)*scale_, t_(0),
                         R_(1, 0)*scale_, R_(1, 1)*scale_, R_(1, 2)*scale_, t_(1),
                         R_(2, 0)*scale_, R_(2, 1)*scale_, R_(2, 2)*scale_, t_(2));
	const Aff_transformation R(R_(0, 0), R_(0, 1), R_(0, 2),
                         R_(1, 0), R_(1, 1), R_(1, 2),
                         R_(2, 0), R_(2, 1), R_(2, 2));
	for (Point_with_normal& pwn: points) {
		pwn = Point_with_normal(T(pwn.first), R(pwn.second));
	}
}

void transform(const Eigen::Matrix3d& R_, std::vector<Vector>& vectors) 
{
	const Aff_transformation R(R_(0, 0), R_(0, 1), R_(0, 2),
                         R_(1, 0), R_(1, 1), R_(1, 2),
                         R_(2, 0), R_(2, 1), R_(2, 2));
	for (Vector& v: vectors) {
		v = R(v);
	}
}

void transform(const Eigen::Matrix3d& R_, const Eigen::Vector3d& t_, const double& scale_, std::vector< std::vector<Point3>>& alpha_triangles) 
{
	const Aff_transformation T(R_(0, 0)*scale_, R_(0, 1)*scale_, R_(0, 2)*scale_, t_(0),
                         R_(1, 0)*scale_, R_(1, 1)*scale_, R_(1, 2)*scale_, t_(1),
                         R_(2, 0)*scale_, R_(2, 1)*scale_, R_(2, 2)*scale_, t_(2));
	for (std::vector<Point3>& triangle: alpha_triangles) {
		triangle[0] = T(triangle[0]);
		triangle[1] = T(triangle[1]);
		triangle[2] = T(triangle[2]);
	}
}

void transform(const Aff_transformation& T, Mesh& mesh) {
	CGAL::Polygon_mesh_processing::transform(T, mesh);
}

double bbox_diag(Bbox& bbox)
{
	double dx = bbox.xmax() - bbox.xmin();
	double dy = bbox.ymax() - bbox.ymin();
	double dz = bbox.zmax() - bbox.zmin();
	return std::sqrt(dx*dx + dy*dy + dz*dz);
}

int main(int argc, char *argv[])
{	
	Parameters params;
	Args args(&params);
	if (!args.parse(argc, argv)) {
		return 1;
	}

    boost::filesystem::path path_processed_pcd(params.path_input_pcd);
    std::string basename = path_processed_pcd.stem().string();

	std::string save_path = params.path_output_dir + separator + basename;
	create_directories(path(params.path_output_dir));
	create_directories(path(save_path));
	std::string filename_Tpred_txt = save_path + separator + basename + ".txt";
	remove( filename_Tpred_txt );

	// Reads the surface mesh file in mesh.
	if (open(params.path_input_mesh)==EXIT_FAILURE) return EXIT_FAILURE;
	// Reads the point set file in points.
    if (open(params.path_input_pcd)==EXIT_FAILURE) return EXIT_FAILURE;

	// set bbox
	m_mesh_bbox=CGAL::Polygon_mesh_processing::bbox(m_input_mesh);
	// set mesh sdf
	sdf.reset(new MeshDistanceField(&m_mesh_bbox, &m_input_mesh));

    // Compute mesh OBB and get diagonal
    std::array<Point3, 8> obb_mesh_points;
    CGAL::oriented_bounding_box(m_input_mesh, obb_mesh_points,
                                CGAL::parameters::use_convex_hull(true));
	Eigen::Vector3d mesh_center(0., 0., 0.);
	for (auto& pt: obb_mesh_points) mesh_center += Eigen::Vector3d(pt.x(), pt.y(), pt.z());
	mesh_center /= 8;
	double diagonal_obb_mesh = 2.0*std::sqrt(CGAL::squared_distance(Point3(mesh_center.x(), mesh_center.y(), mesh_center.z()), obb_mesh_points[0]));


	// Extract primitives from point cloud
	std::cout << "Extract primitives" << std::endl;
	if (m_detected_alpha_shapes.empty()) {
		double spacing(0);
		Shape_detection::detect_shape(m_input_points, params, m_input_points_filtered, m_detected_alpha_shapes, m_normal_alpha_shape, spacing, false);
		if (m_input_points_filtered.size() < m_input_points.size()/5) {
			std::cout << "Too few inliers detected on the point set. Please try another set of parameters." << std::endl;
			return 1;
		}
	}


	// Subsample point cloud
	Pwn_vector sample_points;
	sample_points.reserve(params.nbq);
	if (m_input_points.size()>params.nbq) {
		std::cout << "Subsample point cloud... ";
		std::sample(m_input_points.begin(), m_input_points.end(), std::back_inserter(sample_points),
                params.nbq, std::mt19937{std::random_device{}()});
		std::cout << "done!" << std::endl;
	} else {
		sample_points = m_input_points;
	}

	/* REGISTRATION */
	// Copy
	Pwn_vector input_points_sample, input_points_filtered;
	input_points_sample.reserve(sample_points.size());
	input_points_filtered.reserve(params.nbq);
	for (auto& pt: sample_points) input_points_sample.emplace_back(pt);

	if (m_input_points_filtered.size()>params.nbq) {
		std::sample(m_input_points_filtered.begin(), m_input_points_filtered.end(), std::back_inserter(input_points_filtered),
				params.nbq, std::mt19937{std::random_device{}()});
	} else {
		input_points_filtered = m_input_points_filtered;
	}

	std::vector< std::vector<Point3> > input_alpha_shapes;
	input_alpha_shapes.reserve(m_detected_alpha_shapes.size());
	for (auto& triangle: m_detected_alpha_shapes) {
		std::vector<Point3> temp = {triangle[0], triangle[1], triangle[2]};
		input_alpha_shapes.emplace_back(temp);
	}
	std::vector<Vector>	input_normal_alpha_shapes = m_normal_alpha_shape;

	// sdf from alpha shapes
	TriangleList alpha_triangles;
	for (size_t i=0; i!=input_alpha_shapes.size(); ++i) {
		Point3& p_a = input_alpha_shapes[i][0];
		Point3& p_b = input_alpha_shapes[i][1];
		Point3& p_c = input_alpha_shapes[i][2];
		alpha_triangles.push_back(Triangle(p_a,p_b,p_c));
	}
	sdf_points.reset(new TriangleDistanceField(&m_points_bbox, &alpha_triangles));

	// Compute pcd OBB and diagonal
	std::vector<Point3> input_points_filtered_pos;
	input_points_filtered_pos.reserve(input_points_filtered.size());
	for (auto& pwn: input_points_filtered) input_points_filtered_pos.emplace_back(pwn.first);
	std::array<Point3, 8> obb_pcd_points;
	CGAL::oriented_bounding_box(input_points_filtered_pos, obb_pcd_points,
								CGAL::parameters::use_convex_hull(true));
	Eigen::Vector3d points_center(0., 0., 0.);
	for (auto& pt: obb_pcd_points) points_center += Eigen::Vector3d(pt.x(), pt.y(), pt.z());
	points_center /= 8;
	double diagonal_obb_pcd = 2.0*std::sqrt(CGAL::squared_distance(Point3(points_center.x(), points_center.y(), points_center.z()), obb_pcd_points[0]));

	// initialize rotation
	std::cout << "Initialize rotation" << std::endl;
	// weighted EM with triangle data
	auto fnormals = m_input_mesh.property_map<Face_index, Vector>("f:normal").first;
	std::vector<std::vector<Point3>> mesh_triangles;
	mesh_triangles.reserve(m_input_mesh.faces().size());
	std::vector<Vector> mesh_triangles_normal;
	mesh_triangles_normal.reserve(m_input_mesh.faces().size());
	for (Face_index f: m_input_mesh.faces()) {
		std::vector<Point3> triangle;
		for(Vertex_index vd : vertices_around_face(m_input_mesh.halfedge(f), m_input_mesh) ) {
			triangle.push_back(m_input_mesh.point(vd));
		}
		mesh_triangles.push_back(triangle);
		mesh_triangles_normal.push_back(fnormals[f]);
	}
	std::vector<Vector> normal_means, mesh_normal_means;
	std::vector<double> normal_stds, mesh_normal_stds;
	std::vector<Eigen::Matrix3d> Rlist = Registration::init_rotation(input_alpha_shapes, input_normal_alpha_shapes, mesh_triangles, mesh_triangles_normal, normal_means, mesh_normal_means, normal_stds, mesh_normal_stds);

	// initialize scale
	double scale_init = 1.;
	if (params.scale_est) scale_init = diagonal_obb_mesh/diagonal_obb_pcd;

	// refinement
	std::cout << "Register points to mesh..." << std::endl;
	std::vector<double> s_out_list(Rlist.size());
	std::vector<Eigen::Matrix3d> R_out_list(Rlist.size());
	std::vector<Eigen::Vector3d> t_out_list(Rlist.size());
	std::vector<double> cost_out_list(Rlist.size(), FLT_MAX);
	#pragma omp parallel for
	for (size_t k=0; k<Rlist.size(); k++) {
		int curr = k;
		Eigen::Matrix3d R = Rlist[curr];
		double scale = scale_init;
		Eigen::Vector3d t = mesh_center - scale*R*points_center;
		Registration registration;
		std::pair<double,double> init_final_costs = registration.compute_Rt(input_points_filtered, *sdf_points, m_input_mesh, *sdf, R, t, scale, params.scale_est);
		cost_out_list[curr] = init_final_costs.second;
		s_out_list[curr] = scale;
		R_out_list[curr] = R;
		t_out_list[curr] = t;
	}

	Eigen::Matrix3d R_best;
	Eigen::Vector3d t_best(0., 0., 0.);
	double scale_best = 1.;
	double cost_min = FLT_MAX;
	for (size_t k=0; k<Rlist.size(); k++) {
		if (cost_out_list[k] < cost_min) {
			cost_min = cost_out_list[k];
			scale_best = s_out_list[k];
			R_best = R_out_list[k];
			t_best = t_out_list[k];
		}
	}

	// Transform
	Aff_transformation T(R_best(0, 0)*scale_best, R_best(0, 1)*scale_best, R_best(0, 2)*scale_best, t_best(0),
						R_best(1, 0)*scale_best, R_best(1, 1)*scale_best, R_best(1, 2)*scale_best, t_best(1),
						R_best(2, 0)*scale_best, R_best(2, 1)*scale_best, R_best(2, 2)*scale_best, t_best(2));
	T = T.inverse();
	Mesh res_mesh = m_input_mesh;
	CGAL::Polygon_mesh_processing::transform(T, res_mesh);

	// Save results
	std::string filename_out_mesh = save_path + separator + basename + "_out_mesh.off";
	std::ofstream out_mesh(filename_out_mesh);
	if (!(out_mesh << res_mesh)) {
		return EXIT_FAILURE;
	}

	// txt
	std::ofstream txtoutput_Tpred;
	txtoutput_Tpred.open(filename_Tpred_txt, std::ofstream::out | std::ofstream::app);
	if (txtoutput_Tpred.is_open()) {
		std::string line = to_string_with_precision(T.m(0,0), 6) + " " 
							+ to_string_with_precision(T.m(0,1), 6) + " " 
							+ to_string_with_precision(T.m(0,2), 6) + " " 
							+ to_string_with_precision(T.m(0,3), 6) + "\n" 
							+ to_string_with_precision(T.m(1,0), 6) + " " 
							+ to_string_with_precision(T.m(1,1), 6) + " " 
							+ to_string_with_precision(T.m(1,2), 6) + " " 
							+ to_string_with_precision(T.m(1,3), 6) + "\n" 
							+ to_string_with_precision(T.m(2,0), 6) + " " 
							+ to_string_with_precision(T.m(2,1), 6) + " " 
							+ to_string_with_precision(T.m(2,2), 6) + " " 
							+ to_string_with_precision(T.m(2,3), 6) + "\n" 
							+ to_string_with_precision(0, 6) + " " 
							+ to_string_with_precision(0, 6) + " " 
							+ to_string_with_precision(0, 6) + " " 
							+ to_string_with_precision(1, 6);
		txtoutput_Tpred << line << std::endl;
		txtoutput_Tpred.close();
	} else std::cout << "Unable to open the txt file" << std::endl;

	return 0;
}

