#define CGAL_DISABLE_ROUNDING_MATH_CHECK

#include "args.h"
#include "utils.h"
#include <algorithm>
#include <CGAL/IO/read_xyz_points.h>
#include <CGAL/IO/write_xyz_points.h>
#include <CGAL/IO/OFF_reader.h>
#include <CGAL/Polygon_mesh_processing/orient_polygon_soup.h>
#include <CGAL/Polygon_mesh_processing/polygon_soup_to_polygon_mesh.h>
#include <CGAL/Polygon_mesh_processing/orientation.h>
#include <CGAL/Polygon_mesh_processing/bbox.h>
#include <CGAL/Bbox_3.h>
#include <CGAL/property_map.h>
#include <CGAL/Polygon_mesh_processing/transform.h>
#include <CGAL/Polygon_mesh_processing/compute_normal.h>
#include <fstream>
#include <sstream>
#include <boost/filesystem.hpp>
#include "registration.h"
#include "primitive_detection.h"
#include "RegionGrowing/Visualization_Tools.h"
using namespace boost::filesystem;


Bbox m_points_bbox;
Pwn_vector m_input_points;
Pwn_vector m_input_points_inlier;
std::vector< std::vector<Point3> > m_detected_alpha_shapes;
std::vector<Vector> m_normal_alpha_shape;
std::unique_ptr<TriangleDistanceField> sdf_points;

Bbox m_points_target_bbox;
Pwn_vector m_input_points_target;
Pwn_vector m_input_points_target_inlier;
std::vector< std::vector<Point3> > m_detected_alpha_shapes_target;
std::vector<Vector> m_normal_alpha_shape_target;
std::unique_ptr<TriangleDistanceField> sdf_points_target;

Pwn_vector m_input_points_corr;
Pwn_vector m_input_points_target_corr;

Mesh m_input_mesh;
bool normals_ready = false;
std::unique_ptr<MeshDistanceField> sdf;


template <typename T>
std::string to_string_with_precision(const T a_value, const int n = 8)
{
    std::ostringstream out;
    out.precision(n);
    out << std::fixed << a_value;
    return out.str();
}

void compute_mesh_normals() 
{
    if (m_input_mesh.is_empty() || normals_ready) {return;}
	std::cout << "Compute mesh face normals" << std::endl;
    normals_ready = true;	
    Mesh::Property_map<Face_index, Vector> fnormals = m_input_mesh.add_property_map<Face_index, Vector>("f:normal", CGAL::NULL_VECTOR).first;
	auto vnormal_add = m_input_mesh.add_property_map<Vertex_index, Vector>("v:normal", CGAL::NULL_VECTOR);
	if (vnormal_add.second) {
		Mesh::Property_map<Vertex_index, Vector> vnormals = vnormal_add.first;
		CGAL::Polygon_mesh_processing::compute_normals(m_input_mesh,
			vnormals,
			fnormals
		); 
	} else {
		CGAL::Polygon_mesh_processing::compute_face_normals(m_input_mesh,
			fnormals
		);
	}
	// debug
	int i=0;
	for (auto n : fnormals) {
		if (n.squared_length()==0) {
			std::cout << "Surface normal is zero for " << i << std::endl;
		}
		++i;
	}
}


Bbox set_bbox(Pwn_vector& points)
{
	double x_min, y_min, z_min, x_max, y_max, z_max;
	x_min = y_min = z_min = FLT_MAX;
	x_max = y_max = z_max = -FLT_MAX;
	for (size_t i = 0 ; i < points.size() ; ++i) {
		const Point3 & pt = points[i].first;
		const double &x = pt.x(), &y = pt.y(), &z = pt.z();
		if (x < x_min) x_min = x;
		if (x > x_max) x_max = x;
		if (y < y_min) y_min = y;
		if (y > y_max) y_max = y;
		if (z < z_min) z_min = z;
		if (z > z_max) z_max = z;
	}
	
	return CGAL::Bbox_3(x_min, y_min, z_min, x_max, y_max, z_max);
}


int open_pcd(std::string filename, Pwn_vector&pcd) 
{
	const boost::filesystem::path filepath = filename;
	std::string ext = filepath.extension().string();
	std::ifstream ifs(filename);
	
	if(!ifs)
	{
		std::cerr << "Unable to open the file " << filename << std::endl;
		return EXIT_FAILURE;
	}
	if(ext == ".xyz" || ext == ".XYZ")
	{	
		pcd.clear();
		if (!CGAL::read_xyz_points(ifs,
		std::back_inserter(pcd),
		CGAL::parameters::point_map(Point_map()).normal_map(Normal_map()))) 
		{
			std::cerr << "Error: cannot read file " << filename << std::endl;
			return EXIT_FAILURE;
		}
	}
	else {
		std::cerr << "Unknow extenstion. Please use .xyz for point set" << std::endl;
		return -1;		
	}
	std::cout << "done. (" << m_input_points.size() << " points)" << std::endl;
	ifs.close();
}


void transform(const Eigen::Matrix3d& R_, const Eigen::Vector3d& t_, const double& scale_, Pwn_vector& points) 
{
    const Aff_transformation T(R_(0, 0)*scale_, R_(0, 1)*scale_, R_(0, 2)*scale_, t_(0),
                         R_(1, 0)*scale_, R_(1, 1)*scale_, R_(1, 2)*scale_, t_(1),
                         R_(2, 0)*scale_, R_(2, 1)*scale_, R_(2, 2)*scale_, t_(2));
	const Aff_transformation R(R_(0, 0), R_(0, 1), R_(0, 2),
                         R_(1, 0), R_(1, 1), R_(1, 2),
                         R_(2, 0), R_(2, 1), R_(2, 2));
	for (Point_with_normal& pwn: points) {
		pwn = Point_with_normal(T(pwn.first), R(pwn.second));
	}
}

void transform(const Eigen::Matrix3d& R_, std::vector<Vector>& vectors) 
{
	const Aff_transformation R(R_(0, 0), R_(0, 1), R_(0, 2),
                         R_(1, 0), R_(1, 1), R_(1, 2),
                         R_(2, 0), R_(2, 1), R_(2, 2));
	for (Vector& v: vectors) {
		v = R(v);
	}
}

void transform(const Eigen::Matrix3d& R_, const Eigen::Vector3d& t_, const double& scale_, std::vector< std::vector<Point3>>& alpha_triangles) 
{
	const Aff_transformation T(R_(0, 0)*scale_, R_(0, 1)*scale_, R_(0, 2)*scale_, t_(0),
                         R_(1, 0)*scale_, R_(1, 1)*scale_, R_(1, 2)*scale_, t_(1),
                         R_(2, 0)*scale_, R_(2, 1)*scale_, R_(2, 2)*scale_, t_(2));
	for (std::vector<Point3>& triangle: alpha_triangles) {
		triangle[0] = T(triangle[0]);
		triangle[1] = T(triangle[1]);
		triangle[2] = T(triangle[2]);
	}
}

void transform(const Aff_transformation& T, Mesh& mesh) {
	CGAL::Polygon_mesh_processing::transform(T, mesh);
}

double bbox_diag(Bbox& bbox)
{
	double dx = bbox.xmax() - bbox.xmin();
	double dy = bbox.ymax() - bbox.ymin();
	double dz = bbox.zmax() - bbox.zmin();
	return std::sqrt(dx*dx + dy*dy + dz*dz);
}

int main(int argc, char *argv[])
{	
	Parameters params;
	Args args(&params);
	if (!args.parse(argc, argv)) {
		return 1;
	}

    boost::filesystem::path path_processed_pcd(params.path_input_source_pcd);
    std::string basename = path_processed_pcd.stem().string();

	std::string save_path = params.path_output_dir;
	create_directories(path(params.path_output_dir));

	m_input_points.clear();	
	m_input_points_inlier.clear();
	m_detected_alpha_shapes.clear();
	m_normal_alpha_shape.clear();
	m_input_points_corr.clear();
	m_input_points_target.clear();
	m_input_points_target_inlier.clear();
	m_detected_alpha_shapes_target.clear();
	m_normal_alpha_shape_target.clear();
	m_input_points_target_corr.clear();
	// Reads the point set files in points.
    open_pcd(params.path_input_source_pcd, m_input_points);
	open_pcd(params.path_input_target_pcd, m_input_points_target);

	// Extract primitives from point cloud
	std::cout << "Extract primitives" << std::endl;
	double spacing(0), spacing_target(0);
	std::shared_ptr<RegionGrowing::Cloud_segmentation> C = Shape_detection::detect_shape(m_input_points, params, m_input_points_inlier, m_detected_alpha_shapes, m_normal_alpha_shape, spacing, false);
	if (m_input_points_inlier.size() < m_input_points.size()/5) {
		std::cout << "Too few inliers detected on the point set. Please try another set of parameters." << std::endl;
		return 1;
	}
	// Extract primitives from point cloud
	std::cout << "Extract primitives" << std::endl;
	std::shared_ptr<RegionGrowing::Cloud_segmentation> C_target = Shape_detection::detect_shape(m_input_points_target, params, m_input_points_target_inlier, m_detected_alpha_shapes_target, m_normal_alpha_shape_target, spacing_target, false);
	if (m_input_points_target_inlier.size() < m_input_points_target.size()/5) {
		std::cout << "Too few inliers detected on the point set. Please try another set of parameters." << std::endl;
		return 1;
	}

	// set bbox
	m_points_bbox=set_bbox(m_input_points);
	m_points_target_bbox=set_bbox(m_input_points_target);
	double diag = bbox_diag(m_points_bbox);
	double diag_target = bbox_diag(m_points_target_bbox);
	std::cout<< "diag, diag_target = " << diag << ", " << diag_target << std::endl;

	Pwn_vector input_points_corr, input_points_target_corr;
	std::vector<Point3> m_points_pos_corr, m_points_pos_target_corr;
	if (m_input_points_inlier.size()>params.nbq) {
		std::cout<< "Downsample source inlier pcd" << std::endl;
		std::sample(m_input_points_inlier.begin(), m_input_points_inlier.end(), std::back_inserter(input_points_corr),
				params.nbq, std::mt19937{std::random_device{}()});
	} else input_points_corr = m_input_points_inlier;
	if (m_input_points_target_inlier.size()>params.nbq) {
		std::cout<< "Downsample source inlier pcd" << std::endl;
		std::sample(m_input_points_target_inlier.begin(), m_input_points_target_inlier.end(), std::back_inserter(input_points_target_corr),
				params.nbq, std::mt19937{std::random_device{}()});
	} else input_points_target_corr = m_input_points_target_inlier;

	/* REGISTRATION */

	// initialize rotation
	std::cout << "Initialize rotation" << std::endl;
	std::vector<Vector> normal_means, normal_means_target;
	std::vector<double> normal_stds, normal_stds_target;
	std::vector<Eigen::Matrix3d> Rlist = Registration::init_rotation(m_detected_alpha_shapes, m_normal_alpha_shape, m_detected_alpha_shapes_target, m_normal_alpha_shape_target, normal_means, normal_means_target, normal_stds, normal_stds_target);

	TriangleList alpha_triangles;
	for (size_t i=0; i!=m_detected_alpha_shapes.size(); ++i) {
		Point3& p_a = m_detected_alpha_shapes[i][0];
		Point3& p_b = m_detected_alpha_shapes[i][1];
		Point3& p_c = m_detected_alpha_shapes[i][2];
		alpha_triangles.push_back(Triangle(p_a,p_b,p_c));
	}
	sdf_points.reset(new TriangleDistanceField(&m_points_bbox, &alpha_triangles));

	TriangleList alpha_triangles_target;
	for (size_t i=0; i!=m_detected_alpha_shapes_target.size(); ++i) {
		Point3& p_a = m_detected_alpha_shapes_target[i][0];
		Point3& p_b = m_detected_alpha_shapes_target[i][1];
		Point3& p_c = m_detected_alpha_shapes_target[i][2];
		alpha_triangles_target.push_back(Triangle(p_a,p_b,p_c));
	}
	sdf_points_target.reset(new TriangleDistanceField(&m_points_target_bbox, &alpha_triangles_target));

	// compute point cloud center
	Eigen::Vector3d points_center(0., 0., 0.), points_target_center(0., 0., 0.);
	for (Point_with_normal& pwn: input_points_corr) {
		points_center += Eigen::Vector3d(pwn.first.x(), pwn.first.y(), pwn.first.z());
	}
	points_center /= input_points_corr.size();
	for (Point_with_normal& pwn: input_points_target_corr) {
		points_target_center += Eigen::Vector3d(pwn.first.x(), pwn.first.y(), pwn.first.z());
	}
	points_target_center /= input_points_target_corr.size();

	// initialize scale
	double scale_init = 1.;
	if (params.scale_est) {
		double points_std = 0, points_target_std = 0;
		for (Point_with_normal& pwn: input_points_corr) points_std += (Eigen::Vector3d(pwn.first.x(), pwn.first.y(), pwn.first.z()) - points_center).squaredNorm();
		points_std = std::sqrt(points_std/input_points_corr.size());
		for (Point_with_normal& pwn: input_points_target_corr) points_target_std += (Eigen::Vector3d(pwn.first.x(), pwn.first.y(), pwn.first.z()) - points_target_center).squaredNorm();
		points_target_std = std::sqrt(points_target_std/input_points_target_corr.size());
		scale_init = points_target_std/points_std;
	}

	// refinement
	std::cout << "Register points to mesh..." << std::endl;
	std::vector<double> s_out_list(Rlist.size());
	std::vector<Eigen::Matrix3d> R_out_list(Rlist.size());
	std::vector<Eigen::Vector3d> t_out_list(Rlist.size());
	std::vector<double> cost_out_list(Rlist.size(), FLT_MAX);
	#pragma omp parallel for
	for (size_t k=0; k<Rlist.size(); k++) {
		int curr = k;
		Eigen::Matrix3d R = Rlist[curr];
		double scale = scale_init;
		Eigen::Vector3d t = points_target_center - scale*R*points_center;
		Registration registration;
		std::pair<double,double> init_final_costs = registration.compute_Rt(input_points_corr, *sdf_points, input_points_target_corr, *sdf_points_target, R, t, scale, params.scale_est);
		cost_out_list[curr] = init_final_costs.second;
		s_out_list[curr] = scale;
		R_out_list[curr] = R;
		t_out_list[curr] = t;
	}
	Eigen::Matrix3d R_best;
	Eigen::Vector3d t_best(0., 0., 0.);
	double scale_best = 1.;
	double cost_min = FLT_MAX;
	for (size_t k=0; k<Rlist.size(); k++) {
		if (cost_out_list[k] < cost_min) {
			cost_min = cost_out_list[k];
			scale_best = s_out_list[k];
			R_best = R_out_list[k];
			t_best = t_out_list[k];
		}
	}

	// Transform
	Aff_transformation T(R_best(0, 0)*scale_best, R_best(0, 1)*scale_best, R_best(0, 2)*scale_best, t_best(0),
						R_best(1, 0)*scale_best, R_best(1, 1)*scale_best, R_best(1, 2)*scale_best, t_best(1),
						R_best(2, 0)*scale_best, R_best(2, 1)*scale_best, R_best(2, 2)*scale_best, t_best(2));

	// txt: transformation matrix
	std::string filename_T_txt = save_path + separator + basename + ".txt";
	std::ofstream txtoutput_T;
	txtoutput_T.open(filename_T_txt, std::ofstream::out | std::ofstream::trunc);
	if (txtoutput_T.is_open()) {
		std::string line = to_string_with_precision(T.m(0,0), 6) + " " 
							+ to_string_with_precision(T.m(0,1), 6) + " " 
							+ to_string_with_precision(T.m(0,2), 6) + " " 
							+ to_string_with_precision(T.m(0,3), 6) + "\n" 
							+ to_string_with_precision(T.m(1,0), 6) + " " 
							+ to_string_with_precision(T.m(1,1), 6) + " " 
							+ to_string_with_precision(T.m(1,2), 6) + " " 
							+ to_string_with_precision(T.m(1,3), 6) + "\n" 
							+ to_string_with_precision(T.m(2,0), 6) + " " 
							+ to_string_with_precision(T.m(2,1), 6) + " " 
							+ to_string_with_precision(T.m(2,2), 6) + " " 
							+ to_string_with_precision(T.m(2,3), 6) + "\n" 
							+ to_string_with_precision(0, 6) + " " 
							+ to_string_with_precision(0, 6) + " " 
							+ to_string_with_precision(0, 6) + " " 
							+ to_string_with_precision(1, 6);
		txtoutput_T << line << std::endl;
		txtoutput_T.close();
	} else std::cout << "Unable to open the txt file" << std::endl;

	return 0;
}

