#include "scene.h"
#include "utils.h"
#include <iostream>
#include <fstream>
#include <algorithm>
#include <QFileInfo>
#include <CGAL/IO/read_xyz_points.h>
#include <CGAL/IO/write_xyz_points.h>
#include <CGAL/IO/write_ply_points.h>
#include <CGAL/Surface_mesh/IO/PLY.h>
#include <CGAL/IO/OFF_reader.h>
#include <CGAL/Polygon_mesh_processing/orient_polygon_soup.h>
#include <CGAL/Polygon_mesh_processing/polygon_soup_to_polygon_mesh.h>
#include <CGAL/Polygon_mesh_processing/orientation.h>
#include <CGAL/property_map.h>
#include <CGAL/Polygon_mesh_processing/compute_normal.h>
#include <CGAL/optimal_bounding_box.h>
#include <CGAL/Polygon_mesh_processing/bbox.h>
#include <CGAL/Polygon_mesh_processing/transform.h>
#include <CGAL/Orthogonal_k_neighbor_search.h>
#include <CGAL/Search_traits_3.h>
#include "registration.h"
#include "primitive_detection.h"

typedef std::array<unsigned char, 3> Color;


// Define how a color should be stored
namespace CGAL {
  template< class F > 
  struct Output_rep< ::Color, F > {
    const ::Color& c;
    static const bool is_specialized = true;
    Output_rep (const ::Color& c) : c(c)
    { }
    std::ostream& operator() (std::ostream& out) const
    {
      if (is_ascii(out))
        out << int(c[0]) << " " << int(c[1]) << " " << int(c[2]);
      else
        out.write(reinterpret_cast<const char*>(&c), sizeof(c));
      return out;
    }
  }; 
}


// constants
const int slow_distance_grid_size = 60;
const int fast_distance_grid_size = 20;

Scene::Scene()
	: m_frame (new ManipulatedFrame())
	, m_grid_size(slow_distance_grid_size)
{
	m_points_bbox = Bbox(0., 0., 0., 0., 0., 0.);	
	m_mesh_bbox = Bbox(0., 0., 0., 0., 0., 0.);	

	// distance function
    m_red_ramp.build_red();
    m_blue_ramp.build_blue();
	m_texture.reset(new Texture(m_grid_size,m_grid_size));

	startTimer(0);
	m_ready_to_cut = false;
	normals_ready = false;
}

Scene::~Scene()
{	
	unload();
}

void Scene::unload() 
{	
	m_texture.reset();
	if (sdf) sdf->clear();
	m_input_points.clear();	
	m_input_points_filtered.clear();
	m_input_mesh.clear();
	m_detected_alpha_shapes.clear();
	m_normal_alpha_shape.clear();
	m_tex_map.clear();
	m_pos_plane_arrows.clear();
	m_pos_plane.clear();
	m_points_bbox = Bbox(0., 0., 0., 0., 0., 0.);	
	m_mesh_bbox = Bbox(0., 0., 0., 0., 0., 0.);	
	normals_ready = false;
	set_input_pcd("");
	set_input_mesh("");
}

void Scene::update_grid_size()
{
    m_grid_size = m_fast_distance ? fast_distance_grid_size
                                  : slow_distance_grid_size;
    m_texture.reset(new Texture(m_grid_size,m_grid_size));
}


void Scene::compute_texture(int i, int j,Color_ramp pos_ramp ,Color_ramp neg_ramp)
{	
    const FT& d00 = sdf->distance_function()[i][j].second;
    // determines grey level
    unsigned int i00 = 255-(unsigned)(255.0 * (double)std::fabs(d00) / sdf->max_distance_function());

    if(d00 > 0.0)
        m_texture->setData(i,j,pos_ramp.r(i00),pos_ramp.g(i00),pos_ramp.b(i00));
    else
        m_texture->setData(i,j,neg_ramp.r(i00),neg_ramp.g(i00),neg_ramp.b(i00));
}

/*
Compute elements for fast signed distance computation while rotating the cutting plane
*/
void Scene::compute_plane_elements() 
{
	if (!m_texture) {
		m_texture.reset(new Texture(m_grid_size,m_grid_size));
	}
	if (!sdf || !sdf->cutting_plane_ready()) return;

	m_tex_map.resize(0);
	m_pos_plane.resize(18);
	m_pos_plane_arrows.resize(m_grid_size*m_grid_size*18);
	float diag = float(bbox_diag());
	float half_diag = 0.5f*diag;
	float ratio_size = diag * 0.0025;
	float arrow_length = diag * 0.015;
    //The cutting plane (without frame transformation)
    {
        m_pos_plane[0]= -half_diag; m_pos_plane[1]=-half_diag; m_pos_plane[2]=0.;
        m_pos_plane[3]= -half_diag; m_pos_plane[4]= half_diag; m_pos_plane[5]=0.;
        m_pos_plane[6]=  half_diag; m_pos_plane[7]= half_diag; m_pos_plane[8]=0.;
        m_pos_plane[9]= -half_diag; m_pos_plane[10]= -half_diag; m_pos_plane[11]=0.;
        m_pos_plane[12]= half_diag; m_pos_plane[13]= half_diag; m_pos_plane[14]= 0.;
        m_pos_plane[15]= half_diag; m_pos_plane[16]= -half_diag; m_pos_plane[17]= 0.;

        //UV Mapping
        m_tex_map.push_back(-0.f);
        m_tex_map.push_back(-0.f);

        m_tex_map.push_back(-0.f);
        m_tex_map.push_back(1.f);

        m_tex_map.push_back(1.f);
        m_tex_map.push_back(1.f);

        m_tex_map.push_back(-0.f);
        m_tex_map.push_back(-0.f);

        m_tex_map.push_back(1.f);
        m_tex_map.push_back(1.f);

        m_tex_map.push_back(1.f);
        m_tex_map.push_back(-0.f);
    }
	// Texture
	for( int i=0 ; i < m_texture->getWidth(); i++ )
	{
		for( int j=0 ; j < m_texture->getHeight() ; j++)
		{
			compute_texture(i,j,m_red_ramp,m_blue_ramp);
		}
	}
	// Gradient field on the cutting plane (without frame transformation)
	const auto plane_gradient_function = sdf->gradient_function();
    const float dx = diag;
    const float dy = diag;
	int offset = 0;
    for(int i=0 ; i<m_grid_size ; ++i)
    {
        for(int j=0 ; j<m_grid_size ; ++j)
        {
            const Point& pos_O = plane_gradient_function[i][j].first;
			const Vector& gradient = plane_gradient_function[i][j].second;
			Point pos_A = pos_O + arrow_length*gradient;
			Vector v_line = abs(gradient.z()) > 0.01 ? Vector(1, 0, 0) : Vector(0, 0, 1);
			Vector v1_line = CGAL::cross_product(v_line, gradient);
			Vector v2_line = CGAL::cross_product(v1_line, gradient);
			Point pos_H = pos_A - ratio_size*gradient;
			Point pos_H1 = pos_H + ratio_size*v1_line;
			Point pos_H2 = pos_H - ratio_size*v1_line;
			Point pos_H3 = pos_H + ratio_size*v2_line;
			Point pos_H4 = pos_H - ratio_size*v2_line;

			m_pos_plane_arrows[offset+0] = float(pos_O.x()); m_pos_plane_arrows[offset+1] = float(pos_O.y()); m_pos_plane_arrows[offset+2] = float(pos_O.z());
			m_pos_plane_arrows[offset+3] = float(pos_A.x()); m_pos_plane_arrows[offset+4] = float(pos_A.y()); m_pos_plane_arrows[offset+5] = float(pos_A.z());
			m_pos_plane_arrows[offset+6] = float(pos_H1.x()); m_pos_plane_arrows[offset+7] = float(pos_H1.y()); m_pos_plane_arrows[offset+8] = float(pos_H1.z());
			m_pos_plane_arrows[offset+9] = float(pos_H2.x()); m_pos_plane_arrows[offset+10] = float(pos_H2.y()); m_pos_plane_arrows[offset+11] = float(pos_H2.z());
			m_pos_plane_arrows[offset+12] = float(pos_H3.x()); m_pos_plane_arrows[offset+13] = float(pos_H3.y()); m_pos_plane_arrows[offset+14] = float(pos_H3.z());
			m_pos_plane_arrows[offset+15] = float(pos_H4.x()); m_pos_plane_arrows[offset+16] = float(pos_H4.y()); m_pos_plane_arrows[offset+17] = float(pos_H4.z());
			offset += 18;
		}
    }
}

Aff_transformation Scene::frame_transformation() const
{
    const ::GLdouble* m = m_frame->matrix();

    // OpenGL matrices are row-major matrices
    return Aff_transformation (m[0], m[4], m[8], m[12],
            m[1], m[5], m[9], m[13],
            m[2], m[6], m[10], m[14]);
}

FT Scene::bbox_diag() const
{
	Bbox bbox = m_points_bbox + m_mesh_bbox;
	double dx = bbox.xmax() - bbox.xmin();
	double dy = bbox.ymax() - bbox.ymin();
	double dz = bbox.zmax() - bbox.zmin();
	return FT(std::sqrt(dx*dx + dy*dy + dz*dz));
}

FT Scene::bbox_diag(Bbox& bbox)
{
	double dx = bbox.xmax() - bbox.xmin();
	double dy = bbox.ymax() - bbox.ymin();
	double dz = bbox.zmax() - bbox.zmin();
	return FT(std::sqrt(dx*dx + dy*dy + dz*dz));
}

void Scene::update_points_bbox()
{
	// normalize point normal
	for (size_t i = 0 ; i < m_input_points.size() ; ++i) {
		const Vector & n = m_input_points[i].second;
		const double &x = n.x(), &y = n.y(), &z = n.z();
		double l = std::max(sqrt(x * x + y * y + z * z) , 1e-8);
		m_input_points[i].second = Vector (x / l, y / l, z / l);
	}

	double x_min, y_min, z_min, x_max, y_max, z_max;
	x_min = y_min = z_min = FLT_MAX;
	x_max = y_max = z_max = -FLT_MAX;
	for (size_t i = 0 ; i < m_input_points.size() ; ++i) {
		const Point & p = m_input_points[i].first;
		const double &x = p.x(), &y = p.y(), &z = p.z();
		if (x < x_min) x_min = x;
		if (x > x_max) x_max = x;
		if (y < y_min) y_min = y;
		if (y > y_max) y_max = y;
		if (z < z_min) z_min = z;
		if (z > z_max) z_max = z;
	}
	m_points_bbox = Bbox(x_min, y_min, z_min, x_max, y_max, z_max);
}

void Scene::update_mesh_bbox() { 
	m_mesh_bbox=CGAL::Polygon_mesh_processing::bbox(m_input_mesh); 
}

int Scene::open(QString filename)
{
	QFileInfo fileinfo(filename);
	std::ifstream ifs(filename.toUtf8());
	if(!ifs || !fileinfo.isFile() || ! fileinfo.isReadable())
	{
		std::cerr << "Unable to open the file." << std::endl;
		return -1;
	}
	if(filename.contains(".xyz", Qt::CaseInsensitive))
	{	
		set_input_pcd(filename.toStdString());
		m_input_points.clear();	
		m_input_points_filtered.clear();
		m_detected_alpha_shapes.clear();
		m_normal_alpha_shape.clear();
		std::cout << "Read point set...";
		if (!CGAL::read_xyz_points(ifs,
      	std::back_inserter(m_input_points),
      	CGAL::parameters::point_map(Point_map()).normal_map(Normal_map()))) 
		{
      		std::cerr << "Error: cannot read file " << filename.toStdString() << std::endl;
      		return EXIT_FAILURE;
  		}
		update_points_bbox();
	}
	else if(filename.contains(".off", Qt::CaseInsensitive)) {
		set_input_mesh(filename.toStdString());
		m_input_mesh.clear();
		normals_ready = false;
		std::cout << "Read surface mesh...";
		std::vector<Point3> points;
		std::vector<std::vector<std::size_t> > polygons;
		if(!CGAL::read_OFF(ifs, points, polygons) || points.empty())
		{
			std::cerr << "Cannot open file " << std::endl;
			return EXIT_FAILURE;
		}
		CGAL::Polygon_mesh_processing::orient_polygon_soup(points, polygons);
		CGAL::Polygon_mesh_processing::polygon_soup_to_polygon_mesh(points, polygons, m_input_mesh);
		if (!CGAL::Polygon_mesh_processing::is_outward_oriented(m_input_mesh)) CGAL::Polygon_mesh_processing::reverse_face_orientations(m_input_mesh);
		if (!CGAL::is_triangle_mesh(m_input_mesh)){
			std::cerr << "Input geometry is not triangulated." << std::endl;
			return EXIT_FAILURE;
		}
		update_mesh_bbox();
		compute_mesh_normals();
		sdf.reset(new MeshDistanceField(&m_mesh_bbox, &m_input_mesh));
	}
	else {
		std::cerr << "Unknow extenstion. Please use .xyz for point set and .off for surface mesh" << std::endl;
		return -1;		
	}
	std::cout << "done. (" << m_input_points.size() << " points, " 
						   << m_input_mesh.vertices().size() << " vertices)" << std::endl;
	ifs.close();

	return 0;
}

void Scene::compute_mesh_normals() 
{
    if (m_input_mesh.is_empty() || normals_ready) {return;}
	std::cout << "Compute mesh face normals" << std::endl;
    normals_ready = true;	
    Mesh::Property_map<Face_index, Vector> fnormals = m_input_mesh.add_property_map<Face_index, Vector>("f:normal", CGAL::NULL_VECTOR).first;
	auto vnormal_add = m_input_mesh.add_property_map<Vertex_index, Vector>("v:normal", CGAL::NULL_VECTOR);
	if (vnormal_add.second) {
		Mesh::Property_map<Vertex_index, Vector> vnormals = vnormal_add.first;
		CGAL::Polygon_mesh_processing::compute_normals(m_input_mesh,
			vnormals,
			fnormals
		); 
	} else {
		CGAL::Polygon_mesh_processing::compute_face_normals(m_input_mesh,
			fnormals
		);
	}
	// debug
	int i=0;
	for (auto n : fnormals) {
		if (n.squared_length()==0) {
			std::cout << "Surface normal is zero for " << i << std::endl;
		}
		++i;
	}
}

void Scene::extract_shapes_from_point_cloud()
{
	std::cout << "Extract primitives" << std::endl;
	double spacing(0);
	Shape_detection::detect_shape(m_input_points, params, m_input_points_filtered, m_detected_alpha_shapes, m_normal_alpha_shape, spacing, false);
	m_color_col_alpha_shape = std::vector<CGAL::Color>(m_detected_alpha_shapes.size());
	for (size_t i=0; i<m_detected_alpha_shapes.size(); i++) {
		int red=(int)floor((double)136*rand()/RAND_MAX)+100;
		int green=(int)floor((double)136*rand()/ RAND_MAX)+100;
		int blue=(int)floor((double)136*rand()/ RAND_MAX)+100;
		CGAL::Color col(std::max(0,std::min(255,red)),std::max(0,std::min(255,green)),std::max(0,std::min(255,blue)),120);
		m_color_col_alpha_shape[i] = col;
	}
	if (m_input_points_filtered.size() < m_input_points.size()/5) {
		std::cout << "Too few inliers detected on the point set. Please try another set of parameters." << std::endl;
	}
}


void Scene::compute_distance_function()
{
	compute_mesh_normals();
	Aff_transformation t = frame_transformation();
    return sdf->compute_distance_function(t, m_grid_size);
}

void Scene::register_points_to_polyhedron()
{
	
	if (m_detected_alpha_shapes.empty()) {
		// extract primitives
		extract_shapes_from_point_cloud();
		if (m_detected_alpha_shapes.empty()) {
			std::cout << "No primitives detected on the point set. Please try another set of parameters." << std::endl;
			return;
		}
	}

	Pwn_vector input_points_filtered;
	if (m_input_points_filtered.size()>params.nbq) {
		std::sample(m_input_points_filtered.begin(), m_input_points_filtered.end(), std::back_inserter(input_points_filtered),
				params.nbq, std::mt19937{std::random_device{}()});
	} else {
		input_points_filtered = m_input_points_filtered;
	}

	std::vector< std::vector<Point3> > input_alpha_shapes;
	input_alpha_shapes.reserve(m_detected_alpha_shapes.size());
	for (auto& triangle: m_detected_alpha_shapes) {
		std::vector<Point3> temp = {triangle[0], triangle[1], triangle[2]};
		input_alpha_shapes.emplace_back(temp);
	}
	std::vector<Vector>	input_normal_alpha_shapes = m_normal_alpha_shape;

	// sdf from alpha shapes
	TriangleList alpha_triangles;
	for (size_t i=0; i!=input_alpha_shapes.size(); ++i) {
		Point3& p_a = input_alpha_shapes[i][0];
		Point3& p_b = input_alpha_shapes[i][1];
		Point3& p_c = input_alpha_shapes[i][2];
		alpha_triangles.push_back(Triangle(p_a,p_b,p_c));
	}
	std::unique_ptr<TriangleDistanceField> sdf_points(new TriangleDistanceField(&m_points_bbox, &alpha_triangles));

	// set bbox
	m_mesh_bbox=CGAL::Polygon_mesh_processing::bbox(m_input_mesh);
	// set mesh sdf
	std::unique_ptr<MeshDistanceField> sdf(new MeshDistanceField(&m_mesh_bbox, &m_input_mesh));

    // Compute mesh OBB
    std::array<Point3, 8> obb_mesh_points;
    CGAL::oriented_bounding_box(m_input_mesh, obb_mesh_points,
                                CGAL::parameters::use_convex_hull(true));
	Eigen::Vector3d mesh_center(0., 0., 0.);
	for (auto& pt: obb_mesh_points) mesh_center += Eigen::Vector3d(pt.x(), pt.y(), pt.z());
	mesh_center /= 8;
	double diagonal_obb_mesh = 2.0*std::sqrt(CGAL::squared_distance(Point3(mesh_center.x(), mesh_center.y(), mesh_center.z()), obb_mesh_points[0]));

	// Compute pcd OBB
	std::vector<Point3> input_points_filtered_pos;
	input_points_filtered_pos.reserve(input_points_filtered.size());
	for (auto& pwn: input_points_filtered) input_points_filtered_pos.emplace_back(pwn.first);
	std::array<Point3, 8> obb_pcd_points;
	CGAL::oriented_bounding_box(input_points_filtered_pos, obb_pcd_points,
								CGAL::parameters::use_convex_hull(true));
	Eigen::Vector3d points_center(0., 0., 0.);
	for (auto& pt: obb_pcd_points) points_center += Eigen::Vector3d(pt.x(), pt.y(), pt.z());
	points_center /= 8;
	double diagonal_obb_pcd = 2.0*std::sqrt(CGAL::squared_distance(Point3(points_center.x(), points_center.y(), points_center.z()), obb_pcd_points[0]));

	// initialize rotation
	std::cout << "Initialize rotation" << std::endl;
	// weighted EM with triangle data
	auto fnormals = m_input_mesh.property_map<Face_index, Vector>("f:normal").first;
	std::vector<std::vector<Point3>> mesh_triangles;
	mesh_triangles.reserve(m_input_mesh.faces().size());
	std::vector<Vector> mesh_triangles_normal;
	mesh_triangles_normal.reserve(m_input_mesh.faces().size());
	for (Face_index f: m_input_mesh.faces()) {
		std::vector<Point3> triangle;
		for(Vertex_index vd : vertices_around_face(m_input_mesh.halfedge(f), m_input_mesh) ) {
			triangle.push_back(m_input_mesh.point(vd));
		}
		mesh_triangles.push_back(triangle);
		mesh_triangles_normal.push_back(fnormals[f]);
	}
	std::vector<Vector> normal_means, mesh_normal_means;
	std::vector<double> normal_stds, mesh_normal_stds;
	std::vector<Eigen::Matrix3d> Rlist = Registration::init_rotation(input_alpha_shapes, input_normal_alpha_shapes, mesh_triangles, mesh_triangles_normal, normal_means, mesh_normal_means, normal_stds, mesh_normal_stds);

	// initialize scale
	double scale_init = 1.;
	if (params.scale_est) scale_init = diagonal_obb_mesh/diagonal_obb_pcd;

	// refinement
	std::cout << "Register points to mesh..." << std::endl;
	std::vector<double> s_out_list(Rlist.size());
	std::vector<Eigen::Matrix3d> R_out_list(Rlist.size());
	std::vector<Eigen::Vector3d> t_out_list(Rlist.size());
	std::vector<double> cost_out_list(Rlist.size(), FLT_MAX);
	#pragma omp parallel for
	for (size_t k=0; k<Rlist.size(); k++) {
		int curr = k;
		Eigen::Matrix3d R = Rlist[curr];
		double scale = scale_init;
		Eigen::Vector3d t = mesh_center - scale*R*points_center;
		Registration registration;
		std::pair<double,double> init_final_costs = registration.compute_Rt(input_points_filtered, *sdf_points, m_input_mesh, *sdf, R, t, scale, params.scale_est);
		cost_out_list[curr] = init_final_costs.second;
		s_out_list[curr] = scale;
		R_out_list[curr] = R;
		t_out_list[curr] = t;
	}

	Eigen::Matrix3d R_best;
	Eigen::Vector3d t_best(0., 0., 0.);
	double scale_best = 1.;
	double cost_min = FLT_MAX;
	for (size_t k=0; k<Rlist.size(); k++) {
		if (cost_out_list[k] < cost_min) {
			cost_min = cost_out_list[k];
			scale_best = s_out_list[k];
			R_best = R_out_list[k];
			t_best = t_out_list[k];
		}
	}

	// Transform
	Aff_transformation T(R_best(0, 0)*scale_best, R_best(0, 1)*scale_best, R_best(0, 2)*scale_best, t_best(0),
						R_best(1, 0)*scale_best, R_best(1, 1)*scale_best, R_best(1, 2)*scale_best, t_best(1),
						R_best(2, 0)*scale_best, R_best(2, 1)*scale_best, R_best(2, 2)*scale_best, t_best(2));
	T = T.inverse();
	CGAL::Polygon_mesh_processing::transform(T, m_input_mesh);
}

void Scene::transform(const Eigen::Matrix3d& R_, const Eigen::Vector3d& t_, const double& scale_, Pwn_vector& points) 
{
    const Aff_transformation T(R_(0, 0)*scale_, R_(0, 1)*scale_, R_(0, 2)*scale_, t_(0),
                         R_(1, 0)*scale_, R_(1, 1)*scale_, R_(1, 2)*scale_, t_(1),
                         R_(2, 0)*scale_, R_(2, 1)*scale_, R_(2, 2)*scale_, t_(2));
	const Aff_transformation R(R_(0, 0), R_(0, 1), R_(0, 2),
                         R_(1, 0), R_(1, 1), R_(1, 2),
                         R_(2, 0), R_(2, 1), R_(2, 2));
	for (Point_with_normal& pwn: points) {
		pwn = Point_with_normal(T(pwn.first), R(pwn.second));
	}
}

void Scene::transform(const Eigen::Matrix3d& R_, const Eigen::Vector3d& t_, const double& scale_, std::vector< std::vector<Point3>>& alpha_triangles) 
{
	const Aff_transformation T(R_(0, 0)*scale_, R_(0, 1)*scale_, R_(0, 2)*scale_, t_(0),
                         R_(1, 0)*scale_, R_(1, 1)*scale_, R_(1, 2)*scale_, t_(1),
                         R_(2, 0)*scale_, R_(2, 1)*scale_, R_(2, 2)*scale_, t_(2));
	for (std::vector<Point3>& triangle: alpha_triangles) {
		triangle[0] = T(triangle[0]);
		triangle[1] = T(triangle[1]);
		triangle[2] = T(triangle[2]);
	}
}

void Scene::transform(const Aff_transformation& T, std::vector< std::vector<Point3>>& alpha_triangles) 
{
	for (std::vector<Point3>& triangle: alpha_triangles) {
		triangle[0] = T(triangle[0]);
		triangle[1] = T(triangle[1]);
		triangle[2] = T(triangle[2]);
	}
}

void Scene::transform(const Aff_transformation& T, Mesh& mesh) 
{
	CGAL::Polygon_mesh_processing::transform(T, mesh);
}

void Scene::transform_points() 
{
	Eigen::Matrix3d R;
	R = Eigen::AngleAxisd(params.transform_x_angle, Eigen::Vector3d::UnitX())
		* Eigen::AngleAxisd(params.transform_y_angle, Eigen::Vector3d::UnitY())
		* Eigen::AngleAxisd(params.transform_z_angle, Eigen::Vector3d::UnitZ());
	Eigen::Vector3d t(params.transform_x_offset, params.transform_y_offset, params.transform_z_offset);

	transform(R, t, params.transform_scale, m_input_points);
	if (!m_input_points_filtered.empty()) {
		transform(R, t, params.transform_scale, m_input_points_filtered);
	}
	if (!m_detected_alpha_shapes.empty()) {
		transform(R, t, params.transform_scale, m_detected_alpha_shapes);
	}
}

void Scene::timerEvent(QTimerEvent *)
{
    if(m_frame->isSpinning()) set_fast_distance(true);
    m_ready_to_cut = true;
}