#ifndef _SCENE_
#define _SCENE_

#include <CGAL/Qt/qglviewer.h>
#include <QString>
#include "types.h"
#include "color_ramp.h"
#include "parameters.h"
#include "distance_field.h"
#undef CGAL_CFG_NO_CPP0X_VARIADIC_TEMPLATES

class Texture{
private:
	typedef unsigned char GLubyte;
private:
     int Width;
     int Height;
     int size;
     GLubyte *data;
public:
    Texture(int w, int h)
    {
        Width = w;
        Height = h;
        size = 4*Height*Width;
        data = new GLubyte[Height*Width*4];
    }
	~Texture(){ delete data; }

    int getWidth() const {return Width;}
    int getHeight() const {return Height;}
    int getSize() const {return size;}
    void setData(int i, int j, int r, int g, int b){
        data[4*(Width*j+i) + 0] = r;
        data[4*(Width*j+i) + 1] = g;
        data[4*(Width*j+i) + 2] = b;
		data[4*(Width*j+i) + 3] = 150;
    }

    GLubyte* getData(){return data; }
};


class Scene : public QObject
{
	Q_OBJECT
public: 
	Scene();
	virtual ~Scene();

private:
	typedef CGAL::qglviewer::ManipulatedFrame 					  ManipulatedFrame;
	typedef Point3      Point;
	typedef boost::graph_traits<Mesh>::vertex_descriptor      	  Vertex_Descriptor;
    typedef boost::graph_traits<Mesh>::face_descriptor            Face_Descriptor;
private:
	// member data
	Bbox m_points_bbox;
	Bbox m_mesh_bbox;
	Pwn_vector m_input_points;
	Pwn_vector m_input_points_filtered;
	Mesh m_input_mesh;

	// generated data
	std::vector< std::vector<Point3> > m_detected_alpha_shapes;
	std::vector<Vector> m_normal_alpha_shape;
	std::vector<CGAL::Color> m_color_col_alpha_shape;

	// normals
    bool normals_ready;

	std::vector<float> m_pos_plane; //two triangles
	std::vector<float> m_tex_map;
	std::vector<float> m_pos_plane_arrows;

	// distance functions (simple 2D arrays)
    std::unique_ptr<MeshDistanceField> sdf;
	Color_ramp m_red_ramp;
    Color_ramp m_blue_ramp;
    Color_ramp m_thermal_ramp;

private:
	// frame
	ManipulatedFrame* m_frame;
	int m_grid_size;
	bool m_fast_distance;
	std::shared_ptr<Texture> m_texture;
	void compute_texture(int, int, Color_ramp, Color_ramp);

	bool m_ready_to_cut;

	Aff_transformation frame_transformation() const;
	void update_grid_size();

	// parameters
	Parameters params;
public: 
	// bbox
	const Bbox& points_bbox() const { return m_points_bbox; }
	const Bbox& mesh_bbox() const { return m_mesh_bbox; }
	FT bbox_diag() const;
	FT bbox_diag(Bbox& bbox);
	void update_points_bbox();
	void update_mesh_bbox();
	
	// frame
	ManipulatedFrame* manipulatedFrame() const { return m_frame; }
	void compute_plane_elements();

	// timer sends a top when all the events are finished
    void timerEvent(QTimerEvent *);
	bool ready_to_cut() { return m_ready_to_cut; }
	void ready_to_cut_off() { m_ready_to_cut = false; }

	// parameters
	void set_input_pcd(std::string filepath) { params.path_input_pcd = filepath; }
	void set_input_mesh(std::string filepath) { params.path_input_mesh = filepath; }

	void set_detection_epsilon_ratio(double epsilon_ratio) { params.detection_epsilon_ratio = epsilon_ratio; }
	void set_detection_normal_threshold(double thresh) { params.detection_normal_threshold = thresh; }
	void set_detection_knn(int knn) { params.detection_knn = knn; }
	void set_detection_min_points(int num) { params.detection_min_points = num; }

	void set_x_angle(double angle) { params.transform_x_angle = angle/180.*M_PI; }
	void set_y_angle(double angle) { params.transform_y_angle = angle/180.*M_PI; }
	void set_z_angle(double angle) { params.transform_z_angle = angle/180.*M_PI; }
	void set_x_offset(double offset) { params.transform_x_offset = offset; }
	void set_y_offset(double offset) { params.transform_y_offset = offset; }
	void set_z_offset(double offset) { params.transform_z_offset = offset; }
	void set_scale(double scale) { params.transform_scale = scale; }

	// primitives
	const Pwn_vector &input_points() const { return m_input_points; }
	const Mesh &input_mesh() const { return m_input_mesh; }
	int num_input_points() const { return m_input_points.size(); }
	int num_input_faces() const { return m_input_mesh.number_of_faces(); }
	int num_plane_points() const { return m_pos_plane.size()/3; }
	int num_plane_arrows() const { return m_pos_plane_arrows.size()/18; }
	// alpha_shapes
	const std::vector<std::vector<Point3>>& alpha_shapes() const {return m_detected_alpha_shapes;}
	const std::vector<CGAL::Color>& color_col_alpha_shape() const {return m_color_col_alpha_shape;}
	int num_alpha_shape_faces() const { return m_detected_alpha_shapes.size();}

	std::vector<float> &pos_plane() { return m_pos_plane; }
	std::vector<float> &tex_map() { return m_tex_map; }
	std::vector<float> &pos_plane_arrows() { return m_pos_plane_arrows; }
	std::shared_ptr<Texture> &texture() { return m_texture; }

	// file menu
	int open(QString filename);
	void unload();

	// fast distance setter
    void set_fast_distance(bool b) { m_fast_distance = b; update_grid_size(); }

	// algorithm menu
	void extract_shapes_from_point_cloud();
	std::vector<Eigen::Matrix3d> init_rotation();
	void compute_distance_function();
	void register_points_to_polyhedron();

	// utils
	void transform_points();
	void transform(const Aff_transformation& T, Mesh& mesh);
	void transform(const Aff_transformation& T, std::vector< std::vector<Point3>>& alpha_triangles);
	void transform(const Eigen::Matrix3d& R_, const Eigen::Vector3d& t_, const double& scale_, std::vector< std::vector<Point3>>& alpha_triangles);
	void transform(const Eigen::Matrix3d& R_, const Eigen::Vector3d& t_, const double& scale_, Pwn_vector& points);
	void compute_mesh_normals();
};

#endif // _SCENE_
