#include "scene.h"
#include "viewer.h"

#include <QMouseEvent>
#include <QApplication>
#include <QGLFunctions>
#include <CGAL/Qt/CreateOpenGLContext.h>
#include <random>
#include <map>
#include <Eigen/Eigen>

Viewer::Viewer(QWidget* parent) : 
	CGAL::QGLViewer(parent),
	m_pScene(NULL),
    m_custom_mouse(false)
{
    // view options (keep consistent to ui checkboxes)
	m_view_input_points = true;
	m_view_input_poyhedron = true;
	m_view_plane_sdf = true;
	m_view_plane_gradient = true;
    m_view_extracted_shapes = true;

	setAxisIsDrawn(true);
    are_buffers_initialized = false;
}

Viewer::~Viewer()
{
    m_vao[0].destroy();
    m_vao[1].destroy();
	m_vao[2].destroy();
    m_vao[3].destroy();
    m_vao[4].destroy();

    m_vbo[0].destroy();
    m_vbo[1].destroy();
    m_vbo[2].destroy();
	m_vbo[3].destroy();
    m_vbo[4].destroy();
    m_vbo[5].destroy();
    m_vbo[6].destroy();
    m_vbo[7].destroy();
    m_vbo[8].destroy();
    m_vbo[9].destroy();
    m_vbo[10].destroy();
    m_vbo[11].destroy();
    m_vbo[12].destroy();
}

void Viewer::changed()
{
    m_pScene->compute_plane_elements();
    are_buffers_initialized = false;
}

void Viewer::activate_cutting_plane()
{
    connect(manipulatedFrame(), SIGNAL(modified()), this, SLOT(cutting_plane()));
}

void Viewer::deactivate_cutting_plane()
{
    disconnect(manipulatedFrame(), SIGNAL(modified()), this, SLOT(cutting_plane()));
}

void Viewer::cutting_plane(bool override)
{
    if (m_pScene->ready_to_cut() || override) {
        m_pScene->ready_to_cut_off();
        m_pScene->compute_distance_function();
        changed();
    }
}

Plane Viewer::frame_plane() const
{
    const CGAL::qglviewer::Vec& pos = m_pScene->manipulatedFrame()->position();
    const CGAL::qglviewer::Vec& n = m_pScene->manipulatedFrame()->inverseTransformOf(CGAL::qglviewer::Vec(0.f, 0.f, 1.f));

    return Plane(n[0], n[1],  n[2], - n * pos);
}

Aff_transformation Viewer::frame_transformation() const
{
    const ::GLdouble* m = manipulatedFrame()->matrix();

    // OpenGL matrices are row-major matrices
    return Aff_transformation (m[0], m[4], m[8], m[12],
            m[1], m[5], m[9], m[13],
            m[2], m[6], m[10], m[14]);
}

void Viewer::initializeGL()
{
	CGAL::QGLViewer::initializeGL();
	setBackgroundColor(::Qt::white);
	::glGenTextures(1, &textureId);
    m_vao[0].create(); 
	m_vao[1].create();
	m_vao[2].create();
    m_vao[3].create();
    m_vao[4].create();

    m_vbo[0].create(); 
	m_vbo[1].create(); 
	m_vbo[2].create(); 
	m_vbo[3].create(); 
	m_vbo[4].create(); 
    m_vbo[5].create(); 
    m_vbo[6].create();
    m_vbo[7].create();
    m_vbo[8].create();
    m_vbo[9].create();
    m_vbo[10].create();
    m_vbo[11].create();
    m_vbo[12].create();
    m_vbo[13].create();

    const char vertex_source[] =
    {
        "#version 120 																									\n"
        "attribute highp vec4 vertex;                                                                                   \n"
		"attribute highp vec3 normal;																					\n"
        "attribute vec3 color;																					        \n"
        "varying vec3 vColor; 							                                                                \n"
        "uniform highp mat4 mvp_matrix;	                                                                                \n"
        "uniform highp mat4 f_matrix;																				    \n"
        "uniform highp float point_size;																				\n"
		"uniform bool use_diffuse;																						\n"
        "uniform highp mat4 mv_matrix;																					\n"
		"uniform vec3 camera_position_worldspace;																		\n"
		"varying vec3 light_direction_cameraspace; 																		\n"
		"varying vec3 normal_cameraspace; 																				\n"
        "void main(void) 																								\n"
        "{																												\n"
        "   gl_PointSize = point_size; 																					\n"
        "   gl_Position = mvp_matrix * f_matrix * vertex;																\n"
        "   vColor = color;																                                \n"
        "   if(use_diffuse) {																							\n"
		"   	light_direction_cameraspace = mat3(mv_matrix) * camera_position_worldspace - (mv_matrix * vertex).xyz; 	\n"
		"   	normal_cameraspace = mat3(mv_matrix) * normal;															\n"
		"	}																											\n"
        "}																												\n"
    };
    const char fragment_source[] =
    {
        "#version 120 											 \n"
        "varying vec3 vColor; 							         \n"
		"uniform bool use_diffuse;								 \n"
        "varying vec3 normal_cameraspace; 						 \n"
        "varying vec3 light_direction_cameraspace; 				 \n"
        "void main(void) { 										 \n"
		"   if(!use_diffuse) { gl_FragColor = vec4(vColor, 1.); } 			 \n"
        "   else { 												 \n"
		"		vec3 n = normalize(normal_cameraspace); 		 \n"
        "   	vec3 l = normalize(light_direction_cameraspace); \n"
        "   	float cos_theta = clamp(dot(n,l), 0,1); 		 \n"
        "   	gl_FragColor = vec4(vColor, 1.) * cos_theta; 	 \n"
		"	}													 \n"
        "} 														 \n"
    };
    QOpenGLShader* vertex_shader = new QOpenGLShader(QOpenGLShader::Vertex);
	QOpenGLShader* fragment_shader = new QOpenGLShader(QOpenGLShader::Fragment);
    vertex_shader->compileSourceCode(vertex_source);
    if (!vertex_shader->compileSourceCode(vertex_source)) std::cerr<<"m_shader: Compiling vertex source FAILED"<<std::endl;
    if (!fragment_shader->compileSourceCode(fragment_source)) std::cerr<<"m_shader: Compiling fragmentsource FAILED"<<std::endl;
    if (!m_shader.addShader(vertex_shader)) std::cerr<<"m_shader: adding vertex shader FAILED"<<std::endl;
    if (!m_shader.addShader(fragment_shader)) std::cerr<<"m_shader: adding fragment shader FAILED"<<std::endl;;
    if (!m_shader.link()) std::cerr<<"m_shader: linking Program FAILED"<<std::endl;
    m_shader.bind();

    const char tex_vertex_source[] =
    {
        "#version 120 \n"
        "attribute highp vec4 vertex;\n"
        "attribute highp vec2 tex_coord; \n"
        "uniform highp mat4 mvp_matrix;\n"
        "uniform highp mat4 f_matrix;\n"
        "varying highp vec2 texc;\n"
        "void main(void)\n"
        "{\n"
        "   gl_Position = mvp_matrix * f_matrix * vertex;\n"
        "   texc = tex_coord;\n"
        "}"
    };
    const char tex_fragment_source[] =
    {
        "#version 120 \n"
        "uniform sampler2D texture;\n"
        "varying highp vec2 texc;\n"
        "void main(void) { \n"
        "gl_FragColor = texture2D(texture, texc.st);\n"
        "} \n"
        "\n"
    };
	QOpenGLShader* tex_vertex_shader = new QOpenGLShader(QOpenGLShader::Vertex);
	QOpenGLShader* tex_fragment_shader = new QOpenGLShader(QOpenGLShader::Fragment);
    tex_vertex_shader->compileSourceCode(tex_vertex_source);
    tex_fragment_shader->compileSourceCode(tex_fragment_source);
    m_tex_shader.addShader(tex_vertex_shader);
    m_tex_shader.addShader(tex_fragment_shader);
    if (!m_tex_shader.link()) std::cerr<<"m_tex_shader: linking Program FAILED"<<std::endl;
    m_tex_shader.bind();
}


void Viewer::update_input_buffers()
{
	this->makeCurrent();
	m_shader.bind();
    // points
	std::vector<float> points_buffer_data, points_normal_buffer_data, points_color_buffer_data;
    points_buffer_data.reserve(m_pScene->num_input_points()*3);
    points_normal_buffer_data.reserve(m_pScene->num_input_points()*3);
    points_color_buffer_data.reserve(m_pScene->num_input_points()*3);
	const Pwn_vector &input_points = m_pScene->input_points();
	for (const Point_with_normal &p : input_points) {
		points_buffer_data.push_back((float)p.first[0]);
		points_buffer_data.push_back((float)p.first[1]);
		points_buffer_data.push_back((float)p.first[2]);
        points_normal_buffer_data.push_back((float)p.second[0]);
        points_normal_buffer_data.push_back((float)p.second[1]);
        points_normal_buffer_data.push_back((float)p.second[2]);
        points_color_buffer_data.push_back((float)0.4196);
        points_color_buffer_data.push_back((float)0.3569);
        points_color_buffer_data.push_back((float)0.5843);
	}

    // detected shapes
    std::vector<float> alpha_shapes_pos_buffer_data;
	std::vector<float> alpha_shapes_normal_buffer_data;
    std::vector<float> alpha_shapes_color_buffer_data;
    alpha_shapes_pos_buffer_data.reserve(m_pScene->num_alpha_shape_faces()*9);
    alpha_shapes_normal_buffer_data.reserve(m_pScene->num_alpha_shape_faces()*9);
    alpha_shapes_color_buffer_data.reserve(m_pScene->num_alpha_shape_faces()*9);
    const std::vector<std::vector<Point3>>& alpha_shapes = m_pScene->alpha_shapes();
    const std::vector<CGAL::Color>& color_col_alpha_shape = m_pScene->color_col_alpha_shape();
    for (size_t i=0; i<alpha_shapes.size(); ++i) 
	{
		const Point &p0 = alpha_shapes[i][0];
		const Point &p1 = alpha_shapes[i][1];
		const Point &p2 = alpha_shapes[i][2];
        const CGAL::Color& col = color_col_alpha_shape[i];
        float r = (int)col.r()/255.;
        float g = (int)col.g()/255.;
        float b = (int)col.b()/255.;
		Vector n = CGAL::cross_product(p2-p0, p1-p0);
		n /= std::sqrt(n*n);
	
		alpha_shapes_pos_buffer_data.push_back(p0.x()); alpha_shapes_pos_buffer_data.push_back(p0.y()); alpha_shapes_pos_buffer_data.push_back(p0.z());
		alpha_shapes_pos_buffer_data.push_back(p1.x()); alpha_shapes_pos_buffer_data.push_back(p1.y()); alpha_shapes_pos_buffer_data.push_back(p1.z());
		alpha_shapes_pos_buffer_data.push_back(p2.x()); alpha_shapes_pos_buffer_data.push_back(p2.y()); alpha_shapes_pos_buffer_data.push_back(p2.z());
		alpha_shapes_normal_buffer_data.push_back(-n.x()); alpha_shapes_normal_buffer_data.push_back(-n.y()); alpha_shapes_normal_buffer_data.push_back(-n.z());
		alpha_shapes_normal_buffer_data.push_back(-n.x()); alpha_shapes_normal_buffer_data.push_back(-n.y()); alpha_shapes_normal_buffer_data.push_back(-n.z());
		alpha_shapes_normal_buffer_data.push_back(-n.x()); alpha_shapes_normal_buffer_data.push_back(-n.y()); alpha_shapes_normal_buffer_data.push_back(-n.z());
        alpha_shapes_color_buffer_data.push_back(r);alpha_shapes_color_buffer_data.push_back(g);alpha_shapes_color_buffer_data.push_back(b);
        alpha_shapes_color_buffer_data.push_back(r);alpha_shapes_color_buffer_data.push_back(g);alpha_shapes_color_buffer_data.push_back(b);
        alpha_shapes_color_buffer_data.push_back(r);alpha_shapes_color_buffer_data.push_back(g);alpha_shapes_color_buffer_data.push_back(b);
    }	

    // mesh
	std::vector<float> faces_pos_buffer_data;
	std::vector<float> faces_normal_buffer_data;
    std::vector<float> faces_color_buffer_data(m_pScene->num_input_faces()*9, 0.7);
    faces_pos_buffer_data.reserve(m_pScene->num_input_faces()*9);
    faces_normal_buffer_data.reserve(m_pScene->num_input_faces()*9);
	const Mesh &input_mesh = m_pScene->input_mesh();
	for (Mesh::Face_index f : input_mesh.faces()) 
	{
		Mesh::Halfedge_index he1 = input_mesh.halfedge(f);
		Mesh::Halfedge_index he2 = input_mesh.next(he1);
		Mesh::Halfedge_index he3 = input_mesh.next(he2);
		const Point &p0 = input_mesh.point(input_mesh.target(he1));
		const Point &p1 = input_mesh.point(input_mesh.target(he2));
		const Point &p2 = input_mesh.point(input_mesh.target(he3));
		Vector n = CGAL::cross_product(p2-p0, p1-p0);
		n /= std::sqrt(n*n);
		
		faces_pos_buffer_data.push_back(p0.x()); faces_pos_buffer_data.push_back(p0.y()); faces_pos_buffer_data.push_back(p0.z());
		faces_pos_buffer_data.push_back(p1.x()); faces_pos_buffer_data.push_back(p1.y()); faces_pos_buffer_data.push_back(p1.z());
		faces_pos_buffer_data.push_back(p2.x()); faces_pos_buffer_data.push_back(p2.y()); faces_pos_buffer_data.push_back(p2.z());
		faces_normal_buffer_data.push_back(-n.x()); faces_normal_buffer_data.push_back(-n.y()); faces_normal_buffer_data.push_back(-n.z());
		faces_normal_buffer_data.push_back(-n.x()); faces_normal_buffer_data.push_back(-n.y()); faces_normal_buffer_data.push_back(-n.z());
		faces_normal_buffer_data.push_back(-n.x()); faces_normal_buffer_data.push_back(-n.y()); faces_normal_buffer_data.push_back(-n.z());
	}	
    // cutting plane
	const std::vector<float>& plane_pos_buffer_data = m_pScene->pos_plane();
	const std::vector<float>& tex_map_buffer_data = m_pScene->tex_map();
    // gradient field on the cutting plane
    std::vector<float>& arrow_pos_buffer_data = m_pScene->pos_plane_arrows();
    std::vector<unsigned short> indices_lines;
    std::vector<unsigned short> indices_triangles;
    indices_lines.reserve(m_pScene->num_plane_arrows()*2);
    indices_triangles.reserve(m_pScene->num_plane_arrows()*54);
    int num_arrows = m_pScene->num_plane_arrows();
    int offset = 0;
    for (int i=0 ; i<num_arrows; ++i) 
    {// each index correspond to one vertex
        indices_lines.push_back(offset+0); indices_lines.push_back(offset+1);
        indices_triangles.push_back(offset+1); indices_triangles.push_back(offset+2); indices_triangles.push_back(offset+4);
        indices_triangles.push_back(offset+1); indices_triangles.push_back(offset+2); indices_triangles.push_back(offset+5);
        indices_triangles.push_back(offset+1); indices_triangles.push_back(offset+3); indices_triangles.push_back(offset+4);
        indices_triangles.push_back(offset+1); indices_triangles.push_back(offset+3); indices_triangles.push_back(offset+5);
        indices_triangles.push_back(offset+2); indices_triangles.push_back(offset+3); indices_triangles.push_back(offset+4);
        indices_triangles.push_back(offset+2); indices_triangles.push_back(offset+3); indices_triangles.push_back(offset+5);
        offset += 6;
    }

    // points
    m_vao[0].bind();
    m_vbo[0].bind();
    m_vbo[0].allocate(points_buffer_data.data(), static_cast<int>(points_buffer_data.size()*sizeof(float)));
    int points_vertexLocation = m_shader.attributeLocation("vertex");
    m_shader.enableAttributeArray(points_vertexLocation);
    m_shader.setAttributeBuffer(points_vertexLocation,GL_FLOAT,0,3);
    m_vbo[0].release();
    m_vbo[8].bind();
    m_vbo[8].allocate(points_normal_buffer_data.data(), static_cast<int>(points_normal_buffer_data.size()*sizeof(float)));
    int points_normalLocation = m_shader.attributeLocation("normal");
    m_shader.enableAttributeArray(points_normalLocation);
    m_shader.setAttributeBuffer(points_normalLocation,GL_FLOAT,0,3);
    m_vbo[8].release();
    m_vbo[13].bind();
    m_vbo[13].allocate(points_color_buffer_data.data(), static_cast<int>(points_color_buffer_data.size()*sizeof(float)));
    int points_colorLocation = m_shader.attributeLocation("color");
    m_shader.enableAttributeArray(points_colorLocation);
    m_shader.setAttributeBuffer(points_colorLocation,GL_FLOAT,0,3);
    m_vbo[13].release();
    m_vao[0].release();	
	
    // detected shapes
    m_vao[4].bind();
    m_vbo[9].bind();
    m_vbo[9].allocate(alpha_shapes_pos_buffer_data.data(), static_cast<int>(alpha_shapes_pos_buffer_data.size()*sizeof(float)));
    int alpha_shapes_pos_vertexLocation = m_shader.attributeLocation("vertex");
    m_shader.enableAttributeArray(alpha_shapes_pos_vertexLocation);
    m_shader.setAttributeBuffer(alpha_shapes_pos_vertexLocation,GL_FLOAT,0,3);
    m_vbo[9].release();

    m_vbo[10].bind();
    m_vbo[10].allocate(alpha_shapes_normal_buffer_data.data(), static_cast<int>(alpha_shapes_normal_buffer_data.size()*sizeof(float)));
    int alpha_shapes_normal_vertexLocation = m_shader.attributeLocation("normal");
    m_shader.enableAttributeArray(alpha_shapes_normal_vertexLocation);
    m_shader.setAttributeBuffer(alpha_shapes_normal_vertexLocation,GL_FLOAT,0,3);
    m_vbo[10].release();

    m_vbo[11].bind();
    m_vbo[11].allocate(alpha_shapes_color_buffer_data.data(), static_cast<int>(alpha_shapes_color_buffer_data.size()*sizeof(float)));
    int alpha_shapes_colorLocation = m_shader.attributeLocation("color");
    m_shader.enableAttributeArray(alpha_shapes_colorLocation);
    m_shader.setAttributeBuffer(alpha_shapes_colorLocation,GL_FLOAT,0,3);
    m_vbo[11].release();
    m_vao[4].release();	

    // mesh
	m_vao[1].bind();
    m_vbo[1].bind();
    m_vbo[1].allocate(faces_pos_buffer_data.data(), static_cast<int>(faces_pos_buffer_data.size()*sizeof(float)));
    int faces_pos_vertexLocation = m_shader.attributeLocation("vertex");
    m_shader.enableAttributeArray(faces_pos_vertexLocation);
    m_shader.setAttributeBuffer(faces_pos_vertexLocation,GL_FLOAT,0,3);
    m_vbo[1].release();	

    m_vbo[2].bind();
    m_vbo[2].allocate(faces_normal_buffer_data.data(), static_cast<int>(faces_normal_buffer_data.size()*sizeof(float)));
    int faces_normal_vertexLocation = m_shader.attributeLocation("normal");
    m_shader.enableAttributeArray(faces_normal_vertexLocation);
    m_shader.setAttributeBuffer(faces_normal_vertexLocation,GL_FLOAT,0,3);
    m_vbo[2].release();
    m_vbo[12].bind();
    m_vbo[12].allocate(faces_color_buffer_data.data(), static_cast<int>(faces_color_buffer_data.size()*sizeof(float)));
    int faces_colorLocation = m_shader.attributeLocation("color");
    m_shader.enableAttributeArray(faces_colorLocation);
    m_shader.setAttributeBuffer(faces_colorLocation,GL_FLOAT,0,3);
    m_vbo[12].release();
    m_vao[1].release();

    // distance function gradient
    m_vao[2].bind();
    m_vbo[3].bind();
    m_vbo[3].allocate(arrow_pos_buffer_data.data(), static_cast<int>(arrow_pos_buffer_data.size()*sizeof(float)));
    int arrow_pos_vertexLocation = m_shader.attributeLocation("vertex");
    m_shader.enableAttributeArray(arrow_pos_vertexLocation);
    m_shader.setAttributeBuffer(arrow_pos_vertexLocation,GL_FLOAT,0,3);
    m_vbo[3].release();
    
    m_vbo[4].bind();
    m_vbo[4].allocate(indices_lines.data(), static_cast<int>(indices_lines.size()*sizeof(unsigned short)));
    m_vbo[4].release();
    m_vbo[5].bind();
    m_vbo[4].allocate(indices_triangles.data(), static_cast<int>(indices_triangles.size()*sizeof(unsigned short)));
    m_vbo[5].release();

    m_vao[2].release();
	m_shader.release();

    // distance function
	m_vao[3].bind();
    m_vbo[6].bind();
    m_vbo[6].allocate(plane_pos_buffer_data.data(), static_cast<int>(plane_pos_buffer_data.size()*sizeof(float)));
    int poly_vertexLocation = m_tex_shader.attributeLocation("vertex");
    m_tex_shader.bind();
	m_tex_shader.enableAttributeArray(poly_vertexLocation);
    m_tex_shader.setAttributeBuffer(poly_vertexLocation,GL_FLOAT,0,3);
    m_vbo[6].release();

    m_vbo[7].bind();
    m_vbo[7].allocate(tex_map_buffer_data.data(), static_cast<int>(tex_map_buffer_data.size()*sizeof(float)));
	int tex_Location = m_tex_shader.attributeLocation("tex_coord");
	m_tex_shader.enableAttributeArray(tex_Location);
    m_tex_shader.setAttributeBuffer(tex_Location,GL_FLOAT,0,2);
    m_vbo[7].release();
    m_tex_shader.release();
    if (m_pScene->texture()) {
        ::glBindTexture(GL_TEXTURE_2D, textureId);
	    ::glTexImage2D(GL_TEXTURE_2D,
                 0,
                 GL_RGBA,
                 m_pScene->texture()->getWidth(),
                 m_pScene->texture()->getHeight(),
                 0,
                 GL_RGBA,
                 GL_UNSIGNED_BYTE,
                 m_pScene->texture()->getData());
        ::glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        ::glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        ::glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE );
        ::glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE );
    }
	m_vao[3].release();

    are_buffers_initialized = true;
}


void Viewer::draw()
{
	CGAL::QGLViewer::draw();
	if (m_pScene == NULL) return;
	if (!are_buffers_initialized) {
        update_input_buffers();
    }
	::glEnable(GL_DEPTH_TEST);
    //::glEnable(GL_BLEND);
	//::glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    QMatrix4x4 mvpMatrix, mvMatrix, fMatrix;
    fMatrix.setToIdentity();
    double mvp[16], mv[16];
    this->camera()->getModelViewProjectionMatrix(mvp);
    this->camera()->getModelViewMatrix(mv);
    for(int i=0; i < 16; i++)
    {
        mvpMatrix.data()[i] = (float)mvp[i];
        mvMatrix.data()[i] = (float)mv[i];
    }
	m_shader.bind();
    m_shader.setUniformValue("mvp_matrix", mvpMatrix);	
    m_shader.setUniformValue("mv_matrix", mvMatrix);


	if(m_view_input_poyhedron) 
	{
	    QVector3D camera_pos(camera()->position()[0],camera()->position()[1], camera()->position()[2]);
		m_shader.setUniformValue("f_matrix", fMatrix);
        m_shader.setUniformValue("use_diffuse",  true);
        m_shader.setUniformValue("camera_position_worldspace", camera_pos);
		m_vao[1].bind();
		::glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		::glDrawArrays(GL_TRIANGLES, 0, static_cast<GLsizei>(m_pScene->num_input_faces()*3));
		m_vao[1].release();		
	}

    if (m_view_extracted_shapes)
    {
        QVector3D camera_pos(camera()->position()[0],camera()->position()[1], camera()->position()[2]);
        m_shader.setUniformValue("f_matrix", fMatrix);
		m_shader.setUniformValue("use_diffuse",  false);
        m_shader.setUniformValue("camera_position_worldspace", camera_pos);
		m_vao[4].bind();
		::glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		::glDrawArrays(GL_TRIANGLES, 0, static_cast<GLsizei>(m_pScene->num_alpha_shape_faces()*3));
		m_vao[4].release();
    }
    if(m_view_input_points) 
	{
        fMatrix.setToIdentity();
        m_shader.setUniformValue("f_matrix", fMatrix);
		m_shader.setUniformValue("use_diffuse",  true);
        m_shader.setUniformValue("color", QColor(95, 75, 139, 255));
		m_shader.setUniformValue("point_size", 2.f); 
		m_vao[0].bind();
		::glEnable(GL_POINT_SMOOTH);
		::glDrawArrays(GL_POINTS, 0, static_cast<GLsizei>(m_pScene->num_input_points()));
		m_vao[0].release();
	}

    if(m_view_plane_gradient) 
	{   
        int num_arrows = m_pScene->num_plane_arrows();

        QVector3D camera_pos(camera()->position()[0],camera()->position()[1], camera()->position()[2]);
        m_shader.setUniformValue("f_matrix", fMatrix);
        m_shader.setUniformValue("use_diffuse",  false);
        m_shader.setUniformValue("color", QColor(11, 102, 35, 255));
        m_vao[2].bind();
        ::glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        ::glLineWidth(2);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_vbo[4].bufferId());
        ::glDrawElements(GL_LINES, static_cast<GLsizei>(num_arrows*2), GL_UNSIGNED_SHORT, (void*)0);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_vbo[5].bufferId());
        ::glDrawElements(GL_TRIANGLES, static_cast<GLsizei>(num_arrows*18), GL_UNSIGNED_SHORT, (void*)0);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
        m_vao[2].release();
	}

	if(m_view_plane_sdf) 
	{
	    ::glActiveTexture(GL_TEXTURE0);
        ::glBindTexture(GL_TEXTURE_2D, textureId);

        for(int i=0; i< 16 ; i++)
            fMatrix.data()[i] =  manipulatedFrame()->matrix()[i];
        m_tex_shader.bind();
    	m_tex_shader.setUniformValue("mvp_matrix", mvpMatrix);
        m_tex_shader.setUniformValue("f_matrix", fMatrix);
        m_vao[3].bind();
		::glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        ::glDrawArrays(GL_TRIANGLES, 0, static_cast<GLsizei>(m_pScene->num_plane_points()));
        m_tex_shader.release();
        m_vao[3].release();
	}
	
    m_shader.release();
}

void Viewer::update_bbox()
{
	const Bbox bbox = m_pScene->points_bbox() + m_pScene->mesh_bbox();
	const double xmin = bbox.xmin();
	const double ymin = bbox.ymin();
	const double zmin = bbox.zmin();
	const double xmax = bbox.xmax();
	const double ymax = bbox.ymax();
	const double zmax = bbox.zmax();
	CGAL::qglviewer::Vec 
		vec_min(xmin, ymin, zmin),
		vec_max(xmax, ymax, zmax);
	this->setSceneBoundingBox(vec_min, vec_max);
    std::cout << "Viewer bbox:" << std::endl << xmin << ", " << xmax << std::endl << ymin << ", " << ymax << std::endl << zmin << ", " << zmax << std::endl;

}

void Viewer::mousePressEvent(QMouseEvent* e)
{
    if ( e->modifiers() == Qt::ControlModifier )
    {
        m_pScene->set_fast_distance(true);
        // Refresh distance function
        cutting_plane(true);
        m_custom_mouse = true;
    }
	CGAL::QGLViewer::mousePressEvent(e);
}

void Viewer::mouseReleaseEvent(QMouseEvent* e)
{
    if ( m_custom_mouse )
    {
        m_pScene->set_fast_distance(false);
        // Recompute distance function
        QApplication::setOverrideCursor(Qt::WaitCursor);
        cutting_plane(true);
        QApplication::restoreOverrideCursor();
        m_custom_mouse = false;
    }
	CGAL::QGLViewer::mouseReleaseEvent(e);
}