#include "window.h"
#include "viewer.h"
#include "dialog_box/dialog_box_shape_detection.h"
#include "dialog_box/dialog_box_transform_points.h"
#include "scene.h"

#include "ui_viewer.h"
#include "ui_dialog_box_shape_detection.h"
#include "ui_dialog_box_transform_points.h"

#include <CGAL/Qt/debug.h>
#include <CGAL/IO/read_xyz_points.h>

#include <QDragEnterEvent>
#include <QDropEvent>
#include <QTextStream>
#include <QFileDialog>
#include <QSettings>
#include <QMimeData>

MainWindow::MainWindow(QWidget* parent)
	: CGAL::Qt::DemosMainWindow(parent)
{
	ui = new Ui::MainWindow;
	ui->setupUi(this);

	// saves some pointers from ui, for latter use.
	m_pViewer = ui->viewer;

	// does not save the state of the viewer 
	m_pViewer->setStateFileName(QString::null);

	// accepts drop events
	setAcceptDrops(true);

	// setups scene
	m_pScene = new Scene();
	m_pViewer->setScene(m_pScene);
	m_pViewer->setManipulatedFrame(m_pScene->manipulatedFrame());

	// connects actionQuit (Ctrl+Q) and qApp->quit()
	connect(ui->actionQuit, SIGNAL(triggered()), this, SLOT(quit()));

	this->addRecentFiles(ui->menuFile, ui->actionQuit);
	connect(this, SIGNAL(openRecentFile(QString)), this, SLOT(open(QString)));
	
	readSettings();
}

MainWindow::~MainWindow()
{
	delete(m_pScene);
	delete ui;
}

void MainWindow::quit()
{
	writeSettings();
	close();
}

void MainWindow::readSettings()
{
	this->readState("MainWindow", Size|State);
}

void MainWindow::writeSettings()
{
	this->writeState("MainWindow");
	std::cerr << "Write setting...done.\n";
}

void MainWindow::setAddKeyFrameKeyboardModifiers(::Qt::KeyboardModifiers m)
{
	m_pViewer->setAddKeyFrameKeyboardModifiers(m);
}

void MainWindow::dragEnterEvent(QDragEnterEvent *event)
{
	if (event->mimeData()->hasFormat("text/uri-list"))
		event->acceptProposedAction();
}

void MainWindow::closeEvent(QCloseEvent *event)
{
	writeSettings();
	event->accept();
}

void MainWindow::dropEvent(QDropEvent *event)
{
	QList<QUrl> urls = event->mimeData()->urls();
	QString filename = urls.begin()->toLocalFile();  // read only one input
	if(!filename.isEmpty())
	{
		QTextStream(stderr) << QString("dropEvent(\"%1\")\n").arg(filename);
		open(filename);
	}
	event->acceptProposedAction();
}

void MainWindow::on_actionLoad_input_file_triggered()
{
	QSettings settings;
	QString directory = settings.value("Open directory",
		QDir::current().dirName()).toString();
	QString filename = 
		QFileDialog::getOpenFileName(this,
		tr("Load mesh..."),
		directory,
		"All files (*)");
	if(!filename.isEmpty()) {
		open(filename);
	}
}

void MainWindow::open(QString filename)
{
	QFileInfo fileinfo(filename);
	if(fileinfo.isFile() && fileinfo.isReadable())
	{
		int index = m_pScene->open(filename);
		if(index >= 0)
		{
			//update gui
			m_pViewer->update_input_buffers();
			m_pViewer->update_bbox();
			m_pViewer->showEntireScene();

			ui->actionView_input_points->setEnabled(true);
			ui->actionView_input_polyhedron->setEnabled(true);
			
			// remember used directory for next time
			QSettings settings;
			settings.setValue("Open directory",
				fileinfo.absoluteDir().absolutePath());
			this->addToRecentFiles(filename);	
		}
	}
}


void MainWindow::on_actionSave_snapshot_triggered()
{
	QApplication::setOverrideCursor(Qt::WaitCursor);
	m_pViewer->saveSnapshot();
	QApplication::restoreOverrideCursor();
}

// copy snapshot to clipboard
void MainWindow::on_actionCopy_snapshot_triggered()
{
	QApplication::setOverrideCursor(Qt::WaitCursor);
	QClipboard *qb = QApplication::clipboard();
	m_pViewer->makeCurrent();
	m_pViewer->raise();
	QImage snapshot = m_pViewer->grabFramebuffer();
	qb->setImage(snapshot);
	QApplication::restoreOverrideCursor();
}

void MainWindow::on_actionView_input_points_triggered()
{
	m_pViewer->toggle_view_input_points();
	m_pViewer->update();
}

void MainWindow::on_actionView_input_polyhedron_triggered()
{
	m_pViewer->toggle_view_input_polyhedron();
	m_pViewer->update();
}

void MainWindow::on_actionView_cut_plane_sdf_triggered() 
{
    m_pViewer->toggle_view_plane_sdf();
    m_pViewer->update();
}

void MainWindow::on_actionView_cut_plane_gradient_triggered() 
{
    m_pViewer->toggle_view_plane_gradient();
    m_pViewer->update();
}

void MainWindow::on_actionView_extracted_shapes_triggered() 
{
    m_pViewer->toggle_view_extracted_shapes();
    m_pViewer->update();
}

void MainWindow::on_actionView_transform_points_triggered() 
{
	DialogBoxTransformPoints dialog_box(this);
	if (dialog_box.exec() == QDialog::Accepted) {
		// Get the options
		m_pScene->set_x_angle(dialog_box.GetXAngle());
		m_pScene->set_y_angle(dialog_box.GetYAngle());
		m_pScene->set_z_angle(dialog_box.GetZAngle());
		m_pScene->set_x_offset(dialog_box.GetXOffset());
		m_pScene->set_y_offset(dialog_box.GetYOffset());
		m_pScene->set_z_offset(dialog_box.GetZOffset());
		m_pScene->set_scale(dialog_box.GetScale());
		m_pScene->transform_points();
		m_pViewer->changed();
		m_pViewer->update();
	}
}

void MainWindow::on_actionUnload_all_triggered()
{
	m_pScene->unload();
	m_pViewer->deactivate_cutting_plane();
	m_pViewer->changed();
  	m_pViewer->update();
}

void MainWindow::on_actionExtract_planes_from_point_cloud_triggered() 
{
	DialogBoxShapeDetection dialog_box(this);
	if (dialog_box.exec() == QDialog::Accepted) {
		// Get the options
		m_pScene->set_detection_epsilon_ratio(dialog_box.GetEpsilonRatio());
		m_pScene->set_detection_normal_threshold(dialog_box.GetNormalThreshold());
		m_pScene->set_detection_knn(dialog_box.GetKNN());
		m_pScene->set_detection_min_points(dialog_box.GetMinPoints());
		QApplication::setOverrideCursor(Qt::WaitCursor);
		m_pScene->extract_shapes_from_point_cloud();
		QApplication::restoreOverrideCursor();
		m_pViewer->changed();
		m_pViewer->update();
	}
}

void MainWindow::on_actionDistance_function_to_facets_triggered() 
{
	QApplication::setOverrideCursor(Qt::WaitCursor);
    m_pViewer->activate_cutting_plane();
	m_pScene->compute_distance_function();
	QApplication::restoreOverrideCursor();
	m_pViewer->changed();
	m_pViewer->update();
}

void MainWindow::on_actionRegister_points_to_poyhedron_triggered() 
{
	QApplication::setOverrideCursor(Qt::WaitCursor);
	m_pScene->register_points_to_polyhedron();
	QApplication::restoreOverrideCursor();
	m_pViewer->changed();
	m_pViewer->update();
}
